# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import math
import numpy as np
from unittest.mock import MagicMock
from affine import Affine
import pytest
from pytest import approx
import pickle

from datacube.utils import geometry
from datacube.utils.geometry import (
    GeoBox,
    CRS,
    CRSMismatchError,
    BoundingBox,
    common_crs,
    multigeom,
    bbox_union,
    crs_units_per_degree,
    decompose_rws,
    affine_from_pts,
    get_scale_at_point,
    native_pix_transform,
    scaled_down_geobox,
    compute_reproject_roi,
    roi_normalise,
    roi_shape,
    split_translation,
    is_affine_st,
    apply_affine,
    compute_axis_overlap,
    roi_is_empty,
    clip_lon180,
    chop_along_antimeridian,
    projected_lon,
    w_,
)
from datacube.utils.geometry._base import (
    _mk_crs_coord,
    bounding_box_in_pixel_domain,
    geobox_intersection_conservative,
    geobox_union_conservative,
    force_2d,
    _align_pix,
    _round_to_res,
    _norm_crs,
    _norm_crs_or_error,
    _make_crs_key,
)
from datacube.testutils.geom import (
    epsg4326,
    epsg3577,
    epsg3857,
    AlbersGS,
    mkA,
    xy_from_gbox,
    xy_norm,
    to_fixed_point,
    from_fixed_point,
    gen_test_image_xy,
    SAMPLE_WKT_WITHOUT_AUTHORITY,
)


def test_pickleable():
    poly = geometry.polygon([(10, 20), (20, 20), (20, 10), (10, 20)], crs=epsg4326)
    pickled = pickle.dumps(poly, pickle.HIGHEST_PROTOCOL)
    unpickled = pickle.loads(pickled)
    assert poly == unpickled


def test_geobox_simple():
    t = geometry.GeoBox(4000, 4000,
                        Affine(0.00025, 0.0, 151.0, 0.0, -0.00025, -29.0),
                        epsg4326)

    expect_lon = np.asarray([151.000125, 151.000375, 151.000625, 151.000875, 151.001125,
                             151.001375, 151.001625, 151.001875, 151.002125, 151.002375])

    expect_lat = np.asarray([-29.000125, -29.000375, -29.000625, -29.000875, -29.001125,
                             -29.001375, -29.001625, -29.001875, -29.002125, -29.002375])
    expect_resolution = np.asarray([-0.00025, 0.00025])

    assert t.coordinates['latitude'].values.shape == (4000,)
    assert t.coordinates['longitude'].values.shape == (4000,)

    np.testing.assert_almost_equal(t.resolution, expect_resolution)
    np.testing.assert_almost_equal(t.coords['latitude'].values[:10], expect_lat)
    np.testing.assert_almost_equal(t.coords['longitude'].values[:10], expect_lon)

    assert (t == "some random thing") is False

    # ensure GeoBox accepts string CRS
    assert isinstance(geometry.GeoBox(4000, 4000,
                                      Affine(0.00025, 0.0, 151.0, 0.0, -0.00025, -29.0),
                                      'epsg:4326').crs, CRS)

    # Check GeoBox class is hashable
    t_copy = GeoBox(t.width, t.height, t.transform, t.crs)
    t_other = GeoBox(t.width+1, t.height, t.transform, t.crs)
    assert t_copy is not t
    assert t == t_copy
    assert len(set([t, t, t_copy])) == 1
    assert len(set([t, t_copy, t_other])) == 2


def test_props():
    crs = epsg4326

    box1 = geometry.box(10, 10, 30, 30, crs=crs)
    assert box1
    assert box1.is_valid
    assert not box1.is_empty
    assert box1.area == 400.0
    assert box1.boundary.length == 80.0
    assert box1.centroid == geometry.point(20, 20, crs)

    triangle = geometry.polygon([(10, 20), (20, 20), (20, 10), (10, 20)], crs=crs)
    assert triangle.boundingbox == geometry.BoundingBox(10, 10, 20, 20)
    assert triangle.envelope.contains(triangle)

    assert box1.length == 80.0

    box1copy = geometry.box(10, 10, 30, 30, crs=crs)
    assert box1 == box1copy
    assert box1.convex_hull == box1copy  # NOTE: this might fail because of point order

    box2 = geometry.box(20, 10, 40, 30, crs=crs)
    assert box1 != box2

    bbox = geometry.BoundingBox(1, 0, 10, 13)
    assert bbox.width == 9
    assert bbox.height == 13
    assert bbox.points == [(1, 0), (1, 13), (10, 0), (10, 13)]

    assert bbox.transform(Affine.identity()) == bbox
    assert bbox.transform(Affine.translation(1, 2)) == geometry.BoundingBox(2, 2, 11, 15)

    pt = geometry.point(3, 4, crs)
    assert pt.json['coordinates'] == (3.0, 4.0)
    assert 'Point' in str(pt)
    assert bool(pt) is True
    assert pt.__nonzero__() is True

    # check "CRS as string is converted to class automatically"
    assert isinstance(geometry.point(3, 4, 'epsg:3857').crs, geometry.CRS)

    # constructor with bad input should raise ValueError
    with pytest.raises(ValueError):
        geometry.Geometry(object())


def test_tests():
    box1 = geometry.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geometry.box(20, 10, 40, 30, crs=epsg4326)
    box3 = geometry.box(30, 10, 50, 30, crs=epsg4326)
    box4 = geometry.box(40, 10, 60, 30, crs=epsg4326)
    minibox = geometry.box(15, 15, 25, 25, crs=epsg4326)

    assert not box1.touches(box2)
    assert box1.touches(box3)
    assert not box1.touches(box4)

    assert box1.intersects(box2)
    assert box1.intersects(box3)
    assert not box1.intersects(box4)

    assert not box1.crosses(box2)
    assert not box1.crosses(box3)
    assert not box1.crosses(box4)

    assert not box1.disjoint(box2)
    assert not box1.disjoint(box3)
    assert box1.disjoint(box4)

    assert box1.contains(minibox)
    assert not box1.contains(box2)
    assert not box1.contains(box3)
    assert not box1.contains(box4)

    assert minibox.within(box1)
    assert not box1.within(box2)
    assert not box1.within(box3)
    assert not box1.within(box4)


def test_ops():
    box1 = geometry.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geometry.box(20, 10, 40, 30, crs=epsg4326)
    box3 = geometry.box(20, 10, 40, 30, crs=epsg4326)
    box4 = geometry.box(40, 10, 60, 30, crs=epsg4326)
    no_box = None

    assert box1 != box2
    assert box2 == box3
    assert box3 != no_box

    union1 = box1.union(box2)
    assert union1.area == 600.0

    with pytest.raises(geometry.CRSMismatchError):
        box1.union(box2.to_crs(epsg3857))

    inter1 = box1.intersection(box2)
    assert bool(inter1)
    assert inter1.area == 200.0

    inter2 = box1.intersection(box4)
    assert not bool(inter2)
    assert inter2.is_empty
    # assert not inter2.is_valid  TODO: what's going on here?

    diff1 = box1.difference(box2)
    assert diff1.area == 200.0

    symdiff1 = box1.symmetric_difference(box2)
    assert symdiff1.area == 400.0

    # test segmented
    line = geometry.line([(0, 0), (0, 5), (10, 5)], epsg4326)
    line2 = line.segmented(2)
    assert line.crs is line2.crs
    assert line.length == line2.length
    assert len(line.coords) < len(line2.coords)
    poly = geometry.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)
    poly2 = poly.segmented(2)
    assert poly.crs is poly2.crs
    assert poly.length == poly2.length
    assert poly.area == poly2.area
    assert len(poly.geom.exterior.coords) < len(poly2.geom.exterior.coords)

    poly2 = poly.exterior.segmented(2)
    assert poly.crs is poly2.crs
    assert poly.length == poly2.length
    assert len(poly.geom.exterior.coords) < len(poly2.geom.coords)

    # test interpolate
    pt = line.interpolate(1)
    assert pt.crs is line.crs
    assert pt.coords[0] == (0, 1)
    assert isinstance(pt.coords, list)

    with pytest.raises(TypeError):
        pt.interpolate(3)

    # test array interface
    assert line.__array_interface__ is not None
    assert np.array(line).shape == (3, 2)

    # test simplify
    poly = geometry.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)
    assert poly.simplify(100) == poly

    # test iteration
    poly_2_parts = geometry.Geometry({
        "type": "MultiPolygon",
        "coordinates": [[[[102.0, 2.0], [103.0, 2.0], [103.0, 3.0], [102.0, 3.0], [102.0, 2.0]]],
                        [[[100.0, 0.0], [101.0, 0.0], [101.0, 1.0], [100.0, 1.0], [100.0, 0.0]],
                         [[100.2, 0.2], [100.8, 0.2], [100.8, 0.8], [100.2, 0.8], [100.2, 0.2]]]]},
        'EPSG:4326')
    pp = list(poly_2_parts)
    assert len(pp) == 2
    assert all(p.crs == poly_2_parts.crs for p in pp)

    # test transform
    assert geometry.point(0, 0, epsg4326).transform(lambda x, y: (x+1, y+2)) == geometry.point(1, 2, epsg4326)

    # test sides
    box = geometry.box(1, 2, 11, 22, epsg4326)
    lines = list(geometry.sides(box))
    assert all(line.crs is epsg4326 for line in lines)
    assert len(lines) == 4
    assert lines[0] == geometry.line([(1, 2), (1, 22)], epsg4326)
    assert lines[1] == geometry.line([(1, 22), (11, 22)], epsg4326)
    assert lines[2] == geometry.line([(11, 22), (11, 2)], epsg4326)
    assert lines[3] == geometry.line([(11, 2), (1, 2)], epsg4326)


def test_geom_split():
    box = geometry.box(0, 0, 10, 30, epsg4326)
    line = geometry.line([(5, 0), (5, 30)], epsg4326)
    bb = list(box.split(line))
    assert len(bb) == 2
    assert box.contains(bb[0] | bb[1])
    assert (box ^ (bb[0] | bb[1])).is_empty

    with pytest.raises(CRSMismatchError):
        list(box.split(geometry.line([(5, 0), (5, 30)], epsg3857)))


def test_common_crs():
    assert common_crs([]) is None
    assert common_crs([geometry.point(0, 0, epsg4326),
                       geometry.line([(0, 0), (1, 1)], epsg4326)]) is epsg4326

    with pytest.raises(CRSMismatchError):
        common_crs([geometry.point(0, 0, epsg4326),
                    geometry.line([(0, 0), (1, 1)], epsg3857)])


def test_multigeom():
    p1, p2 = (0, 0), (1, 2)
    p3, p4 = (3, 4), (5, 6)
    b1 = geometry.box(*p1, *p2, epsg4326)
    b2 = geometry.box(*p3, *p4, epsg4326)
    bb = multigeom([b1, b2])
    assert bb.type == 'MultiPolygon'
    assert bb.crs is b1.crs
    assert len(list(bb)) == 2

    g1 = geometry.line([p1, p2], None)
    g2 = geometry.line([p3, p4], None)
    gg = multigeom(iter([g1, g2, g1]))
    assert gg.type == 'MultiLineString'
    assert gg.crs is g1.crs
    assert len(list(gg)) == 3

    g1 = geometry.point(*p1, epsg3857)
    g2 = geometry.point(*p2, epsg3857)
    g3 = geometry.point(*p3, epsg3857)
    gg = multigeom(iter([g1, g2, g3]))
    assert gg.type == 'MultiPoint'
    assert gg.crs is g1.crs
    assert len(list(gg)) == 3
    assert list(gg)[0] == g1
    assert list(gg)[1] == g2
    assert list(gg)[2] == g3

    # can't mix types
    with pytest.raises(ValueError):
        multigeom([geometry.line([p1, p2], None), geometry.point(*p1, None)])

    # can't mix CRSs
    with pytest.raises(CRSMismatchError):
        multigeom([geometry.line([p1, p2], epsg4326),
                   geometry.line([p3, p4], epsg3857)])

    # only some types are supported on input
    with pytest.raises(ValueError):
        multigeom([gg])


def test_shapely_wrappers():
    poly = geometry.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)

    assert isinstance(poly.svg(), str)
    assert isinstance(poly._repr_svg_(), str)

    with_hole = poly.buffer(1) - poly
    assert len(poly.interiors) == 0
    assert len(with_hole.interiors) == 1
    assert isinstance(with_hole.interiors, list)
    assert isinstance(with_hole.interiors[0], geometry.Geometry)
    assert with_hole.interiors[0].crs == with_hole.crs
    assert poly.exterior.crs == poly.crs

    x, y = poly.exterior.xy
    assert len(x) == len(y)
    assert x.typecode == y.typecode
    assert x.typecode == 'd'

    assert ((poly | poly) ^ poly).is_empty
    assert ((poly & poly) ^ poly).is_empty
    assert (poly ^ poly).is_empty
    assert (poly - poly).is_empty


def test_to_crs():
    poly = geometry.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)
    num_points = 3
    assert poly.crs is epsg4326
    assert poly.to_crs(epsg3857).crs is epsg3857
    assert poly.to_crs('EPSG:3857').crs == 'EPSG:3857'
    assert poly.to_crs('EPSG:3857', 0.1).crs == epsg3857

    assert poly.exterior.to_crs(epsg3857) == poly.to_crs(epsg3857).exterior

    # test that by default segmentation happens
    # +1 is because exterior loops back to start point
    assert len(poly.to_crs(epsg3857).exterior.xy[0]) > num_points + 1

    # test that +inf disables segmentation
    # +1 is because exterior loops back to start point
    assert len(poly.to_crs(epsg3857, float('+inf')).exterior.xy[0]) == num_points + 1

    # test the segmentation works on multi-polygons
    mpoly = (geometry.box(0, 0, 1, 3, 'EPSG:4326') |
             geometry.box(2, 4, 3, 6, 'EPSG:4326'))

    assert mpoly.type == 'MultiPolygon'
    assert mpoly.to_crs(epsg3857).type == 'MultiPolygon'

    poly = geometry.polygon([(0, 0), (0, 5), (10, 5)], None)
    assert poly.crs is None
    with pytest.raises(ValueError):
        poly.to_crs(epsg3857)


def test_boundingbox():
    bb = BoundingBox(0, 3, 2, 4)
    assert bb.width == 2
    assert bb.height == 1
    assert bb.width == bb.span_x
    assert bb.height == bb.span_y

    bb = BoundingBox(0, 3, 2.1, 4)
    assert bb.width == 2
    assert bb.height == 1
    assert bb.span_x == 2.1
    assert bb.width != bb.span_x
    assert bb.height == bb.span_y

    assert BoundingBox.from_xy(bb.range_x, bb.range_y) == bb

    assert BoundingBox.from_xy((1, 2), (10, 20)) == (1, 10, 2, 20)
    assert BoundingBox.from_xy((2, 1), (20, 10)) == (1, 10, 2, 20)
    assert BoundingBox.from_points((1, 11), (2, 22)) == (1, 11, 2, 22)
    assert BoundingBox.from_points((1, 22), (2, 11)) == (1, 11, 2, 22)


def test_densify():
    from datacube.utils.geometry._base import densify

    s_x10 = [(0, 0), (10, 0)]
    assert densify(s_x10, 20) == s_x10
    assert densify(s_x10, 200) == s_x10
    assert densify(s_x10, 5) == [(0, 0), (5, 0), (10, 0)]
    assert densify(s_x10, 4) == [(0, 0), (4, 0), (8, 0), (10, 0)]


def test_bbox_union():
    b1 = BoundingBox(0, 1, 10, 20)
    b2 = BoundingBox(5, 6, 11, 22)

    assert bbox_union([b1]) == b1
    assert bbox_union([b2]) == b2

    bb = bbox_union(iter([b1, b2]))
    assert bb == BoundingBox(0, 1, 11, 22)

    bb = bbox_union(iter([b2, b1]*10))
    assert bb == BoundingBox(0, 1, 11, 22)


def test_unary_union():
    box1 = geometry.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geometry.box(20, 10, 40, 30, crs=epsg4326)
    box3 = geometry.box(30, 10, 50, 30, crs=epsg4326)
    box4 = geometry.box(40, 10, 60, 30, crs=epsg4326)

    union0 = geometry.unary_union([box1])
    assert union0 == box1

    union1 = geometry.unary_union([box1, box4])
    assert union1.type == 'MultiPolygon'
    assert union1.area == 2.0 * box1.area

    union2 = geometry.unary_union([box1, box2])
    assert union2.type == 'Polygon'
    assert union2.area == 1.5 * box1.area

    union3 = geometry.unary_union([box1, box2, box3, box4])
    assert union3.type == 'Polygon'
    assert union3.area == 2.5 * box1.area

    union4 = geometry.unary_union([union1, box2, box3])
    assert union4.type == 'Polygon'
    assert union4.area == 2.5 * box1.area

    assert geometry.unary_union([]) is None

    with pytest.raises(ValueError):
        geometry.unary_union([box1, box1.to_crs(epsg3577)])


def test_unary_intersection():
    box1 = geometry.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geometry.box(15, 10, 35, 30, crs=epsg4326)
    box3 = geometry.box(20, 10, 40, 30, crs=epsg4326)
    box4 = geometry.box(25, 10, 45, 30, crs=epsg4326)
    box5 = geometry.box(30, 10, 50, 30, crs=epsg4326)
    box6 = geometry.box(35, 10, 55, 30, crs=epsg4326)

    inter1 = geometry.unary_intersection([box1])
    assert bool(inter1)
    assert inter1 == box1

    inter2 = geometry.unary_intersection([box1, box2])
    assert bool(inter2)
    assert inter2.area == 300.0

    inter3 = geometry.unary_intersection([box1, box2, box3])
    assert bool(inter3)
    assert inter3.area == 200.0

    inter4 = geometry.unary_intersection([box1, box2, box3, box4])
    assert bool(inter4)
    assert inter4.area == 100.0

    inter5 = geometry.unary_intersection([box1, box2, box3, box4, box5])
    assert bool(inter5)
    assert inter5.type == 'LineString'
    assert inter5.length == 20.0

    inter6 = geometry.unary_intersection([box1, box2, box3, box4, box5, box6])
    assert not bool(inter6)
    assert inter6.is_empty


class TestCRSEqualityComparisons(object):
    def test_comparison_edge_cases(self):
        a = epsg4326
        none_crs = None
        assert a == a
        assert a == str(a)
        assert (a == none_crs) is False
        assert (a == []) is False
        assert (a == TestCRSEqualityComparisons) is False

    def test_australian_albers_comparison(self):
        a = geometry.CRS("""PROJCS["GDA94_Australian_Albers",GEOGCS["GCS_GDA_1994",
                            DATUM["Geocentric_Datum_of_Australia_1994",SPHEROID["GRS_1980",6378137,298.257222101]],
                            PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]],
                            PROJECTION["Albers_Conic_Equal_Area"],
                            PARAMETER["standard_parallel_1",-18],
                            PARAMETER["standard_parallel_2",-36],
                            PARAMETER["latitude_of_center",0],
                            PARAMETER["longitude_of_center",132],
                            PARAMETER["false_easting",0],
                            PARAMETER["false_northing",0],
                            UNIT["Meter",1]]""")
        b = epsg3577

        assert a == b

        assert a != epsg4326


def test_no_epsg():
    c = geometry.CRS('+proj=longlat +no_defs +ellps=GRS80')
    b = geometry.CRS("""GEOGCS["GRS 1980(IUGG, 1980)",DATUM["unknown",SPHEROID["GRS80",6378137,298.257222101]],
                        PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]]""")

    assert c.epsg is None
    assert b.epsg is None


def test_xy_from_geobox():
    gbox = GeoBox(3, 7, Affine.translation(10, 1000), epsg3857)
    xx, yy = xy_from_gbox(gbox)

    assert xx.shape == gbox.shape
    assert yy.shape == gbox.shape
    assert (xx[:, 0] == 10.5).all()
    assert (xx[:, 1] == 11.5).all()
    assert (yy[0, :] == 1000.5).all()
    assert (yy[6, :] == 1006.5).all()

    xx_, yy_, A = xy_norm(xx, yy)
    assert xx_.shape == xx.shape
    assert yy_.shape == yy.shape
    np.testing.assert_almost_equal((xx_.min(), xx_.max()), (0, 1))
    np.testing.assert_almost_equal((yy_.min(), yy_.max()), (0, 1))
    assert (xx_[0] - xx_[1]).sum() != 0
    assert (xx_[:, 0] - xx_[:, 1]).sum() != 0

    XX, YY = apply_affine(A, xx_, yy_)
    np.testing.assert_array_almost_equal(xx, XX)
    np.testing.assert_array_almost_equal(yy, YY)


def test_gen_test_image_xy():
    gbox = GeoBox(3, 7, Affine.translation(10, 1000), epsg3857)

    xy, denorm = gen_test_image_xy(gbox, 'float64')
    assert xy.dtype == 'float64'
    assert xy.shape == (2,) + gbox.shape

    x, y = denorm(xy)
    x_, y_ = xy_from_gbox(gbox)

    np.testing.assert_almost_equal(x, x_)
    np.testing.assert_almost_equal(y, y_)

    xy, denorm = gen_test_image_xy(gbox, 'uint16')
    assert xy.dtype == 'uint16'
    assert xy.shape == (2,) + gbox.shape

    x, y = denorm(xy[0], xy[1])
    assert x.shape == xy.shape[1:]
    assert y.shape == xy.shape[1:]
    assert x.dtype == 'float64'

    x_, y_ = xy_from_gbox(gbox)

    np.testing.assert_almost_equal(x, x_, 4)
    np.testing.assert_almost_equal(y, y_, 4)

    for dt in ('int8', np.int16, np.dtype(np.uint64)):
        xy, _ = gen_test_image_xy(gbox, dt)
        assert xy.dtype == dt

    # check no-data
    xy, denorm = gen_test_image_xy(gbox, 'float32')
    assert xy.dtype == 'float32'
    assert xy.shape == (2,) + gbox.shape
    xy[0, 0, :] = np.nan
    xy[1, 1, :] = np.nan
    xy_ = denorm(xy, nodata=np.nan)
    assert np.isnan(xy_[:, :2]).all()
    np.testing.assert_almost_equal(xy_[0][2:], x_[2:], 6)
    np.testing.assert_almost_equal(xy_[1][2:], y_[2:], 6)

    xy, denorm = gen_test_image_xy(gbox, 'int16')
    assert xy.dtype == 'int16'
    assert xy.shape == (2,) + gbox.shape
    xy[0, 0, :] = -999
    xy[1, 1, :] = -999
    xy_ = denorm(xy, nodata=-999)
    assert np.isnan(xy_[:, :2]).all()
    np.testing.assert_almost_equal(xy_[0][2:], x_[2:], 4)
    np.testing.assert_almost_equal(xy_[1][2:], y_[2:], 4)

    # call without arguments should return linear mapping
    A = denorm()
    assert isinstance(A, Affine)


def test_fixed_point():
    aa = np.asarray([0, 0.5, 1])
    uu = to_fixed_point(aa, 'uint8')
    assert uu.dtype == 'uint8'
    assert aa.shape == uu.shape
    assert tuple(uu.ravel()) == (0, 128, 255)

    aa_ = from_fixed_point(uu)
    assert aa_.dtype == 'float64'
    dd = np.abs(aa - aa_)
    assert (dd < 1/255.0).all()

    uu = to_fixed_point(aa, 'uint16')
    assert uu.dtype == 'uint16'
    assert tuple(uu.ravel()) == (0, 0x8000, 0xFFFF)

    uu = to_fixed_point(aa, 'int16')
    assert uu.dtype == 'int16'
    assert tuple(uu.ravel()) == (0, 0x4000, 0x7FFF)

    aa_ = from_fixed_point(uu)
    dd = np.abs(aa - aa_)
    assert (dd < 1.0/0x7FFF).all()


def test_geobox():
    points_list = [
        [(148.2697, -35.20111), (149.31254, -35.20111), (149.31254, -36.331431), (148.2697, -36.331431)],
        [(148.2697, 35.20111), (149.31254, 35.20111), (149.31254, 36.331431), (148.2697, 36.331431)],
        [(-148.2697, 35.20111), (-149.31254, 35.20111), (-149.31254, 36.331431), (-148.2697, 36.331431)],
        [(-148.2697, -35.20111), (-149.31254, -35.20111), (-149.31254, -36.331431), (-148.2697, -36.331431),
         (148.2697, -35.20111)],
    ]
    for points in points_list:
        polygon = geometry.polygon(points, crs=epsg3577)
        resolution = (-25, 25)
        geobox = geometry.GeoBox.from_geopolygon(polygon, resolution)

        assert abs(resolution[0]) > abs(geobox.extent.boundingbox.left - polygon.boundingbox.left)
        assert abs(resolution[0]) > abs(geobox.extent.boundingbox.right - polygon.boundingbox.right)
        assert abs(resolution[1]) > abs(geobox.extent.boundingbox.top - polygon.boundingbox.top)
        assert abs(resolution[1]) > abs(geobox.extent.boundingbox.bottom - polygon.boundingbox.bottom)

    A = mkA(0, scale=(10, -10),
            translation=(-48800, -2983006))

    w, h = 512, 256
    gbox = GeoBox(w, h, A, epsg3577)

    assert gbox.shape == (h, w)
    assert gbox.transform == A
    assert gbox.extent.crs == gbox.crs
    assert gbox.geographic_extent.crs == epsg4326
    assert gbox.extent.boundingbox.height == h*10.0
    assert gbox.extent.boundingbox.width == w*10.0
    assert isinstance(str(gbox), str)
    assert 'EPSG:3577' in repr(gbox)

    assert GeoBox(1, 1, mkA(0), epsg4326).geographic_extent.crs == epsg4326
    assert GeoBox(1, 1, mkA(0), None).dimensions == ('y', 'x')

    g2 = gbox[:-10, :-20]
    assert g2.shape == (gbox.height - 10, gbox.width - 20)

    # step of 1 is ok
    g2 = gbox[::1, ::1]
    assert g2.shape == gbox.shape

    assert gbox[0].shape == (1, gbox.width)
    assert gbox[:3].shape == (3, gbox.width)

    with pytest.raises(NotImplementedError):
        gbox[::2, :]

    # too many slices
    with pytest.raises(ValueError):
        gbox[:1, :1, :]

    assert gbox.buffered(10, 0).shape == (gbox.height + 2*1, gbox.width)
    assert gbox.buffered(30, 20).shape == (gbox.height + 2*3, gbox.width + 2*2)

    assert (gbox | gbox) == gbox
    assert (gbox & gbox) == gbox
    assert gbox.is_empty() is False
    assert bool(gbox) is True

    assert (gbox[:3, :4] & gbox[3:, 4:]).is_empty()
    assert (gbox[:3, :4] & gbox[30:, 40:]).is_empty()

    with pytest.raises(ValueError):
        geobox_intersection_conservative([])

    with pytest.raises(ValueError):
        geobox_union_conservative([])

    # can not combine across CRSs
    with pytest.raises(ValueError):
        bounding_box_in_pixel_domain(GeoBox(1, 1, mkA(0), epsg4326),
                                     GeoBox(2, 3, mkA(0), epsg3577))


def test_geobox_xr_coords():
    A = mkA(0, scale=(10, -10),
            translation=(-48800, -2983006))

    w, h = 512, 256
    gbox = GeoBox(w, h, A, epsg3577)

    cc = gbox.xr_coords()
    assert list(cc) == ['y', 'x']
    assert cc['y'].shape == (gbox.shape[0],)
    assert cc['x'].shape == (gbox.shape[1],)
    assert 'crs' in cc['y'].attrs
    assert 'crs' in cc['x'].attrs

    cc = gbox.xr_coords(with_crs=True)
    assert list(cc) == ['y', 'x', 'spatial_ref']
    assert cc['spatial_ref'].shape == ()
    assert cc['spatial_ref'].attrs['spatial_ref'] == gbox.crs.wkt
    assert isinstance(cc['spatial_ref'].attrs['grid_mapping_name'], str)

    cc = gbox.xr_coords(with_crs='Albers')
    assert list(cc) == ['y', 'x', 'Albers']

    # geographic CRS
    A = mkA(0, scale=(0.1, -0.1), translation=(10, 30))
    gbox = GeoBox(w, h, A, 'epsg:4326')
    cc = gbox.xr_coords(with_crs=True)
    assert list(cc) == ['latitude', 'longitude', 'spatial_ref']
    assert cc['spatial_ref'].shape == ()
    assert cc['spatial_ref'].attrs['spatial_ref'] == gbox.crs.wkt
    assert isinstance(cc['spatial_ref'].attrs['grid_mapping_name'], str)

    # missing CRS for GeoBox
    gbox = GeoBox(w, h, A, None)
    cc = gbox.xr_coords(with_crs=True)
    assert list(cc) == ['y', 'x']

    # check CRS without name
    crs = MagicMock()
    crs.projected = True
    crs.wkt = epsg3577.wkt
    crs.epsg = epsg3577.epsg
    crs._crs = MagicMock()
    crs._crs.to_cf.return_value = {}
    assert _mk_crs_coord(crs).attrs['grid_mapping_name'] == '??'


def test_projected_lon():
    assert projected_lon(epsg3857, 180).crs is epsg3857
    assert projected_lon('EPSG:3577', 100).crs == epsg3577


def test_chop():
    poly = geometry.box(618300, -1876800,
                        849000, -1642500, 'EPSG:32660')

    chopped = chop_along_antimeridian(poly)
    assert chopped.crs is poly.crs
    assert chopped.type == 'MultiPolygon'
    assert len([g for g in chopped]) == 2

    poly = geometry.box(0, 0, 10, 20, 'EPSG:4326')._to_crs(epsg3857)
    assert poly.crs is epsg3857
    assert chop_along_antimeridian(poly) is poly

    with pytest.raises(ValueError):
        chop_along_antimeridian(geometry.box(0, 1, 2, 3, None))


def test_clip_lon180():
    err = 1e-9

    def b(rside):
        return geometry.box(170, 0, rside, 10, epsg4326)

    def b_neg(lside):
        return geometry.box(lside, 0, -170, 10, epsg4326)

    assert clip_lon180(b(180 - err)) == b(180)
    assert clip_lon180(b(-180 + err)) == b(180)

    assert clip_lon180(b_neg(180 - err)) == b_neg(-180)
    assert clip_lon180(b_neg(-180 + err)) == b_neg(-180)

    bb = multigeom([b(180-err), b_neg(180-err)])
    bb_ = [g for g in clip_lon180(bb)]
    assert bb_[0] == b(180)
    assert bb_[1] == b_neg(-180)


def test_wrap_dateline():
    albers_crs = epsg3577
    geog_crs = epsg4326

    wrap = geometry.polygon([(3658653.1976781483, -4995675.379595791),
                             (4025493.916030875, -3947239.249752495),
                             (4912789.243100313, -4297237.125269571),
                             (4465089.861944263, -5313778.16975072),
                             (3658653.1976781483, -4995675.379595791)], crs=albers_crs)
    wrapped = wrap.to_crs(geog_crs)
    assert wrapped.type == 'Polygon'
    assert wrapped.intersects(geometry.line([(0, -90), (0, 90)], crs=geog_crs))
    wrapped = wrap.to_crs(geog_crs, wrapdateline=True)
    assert wrapped.type == 'MultiPolygon'
    assert not wrapped.intersects(geometry.line([(0, -90), (0, 90)], crs=geog_crs))


@pytest.mark.parametrize("pts", [
    [(12231455.716333, -5559752.598333),
     (12231455.716333, -4447802.078667),
     (13343406.236, -4447802.078667),
     (13343406.236, -5559752.598333),
     (12231455.716333, -5559752.598333)],
    [(13343406.236, -5559752.598333),
     (13343406.236, -4447802.078667),
     (14455356.755667, -4447802.078667),
     (14455356.755667, -5559752.598333),
     (13343406.236, -5559752.598333)],
    [(14455356.755667, -5559752.598333),
     (14455356.755667, -4447802.078667),
     (15567307.275333, -4447802.078667),
     (15567307.275333, -5559752.598333),
     (14455356.755667, -5559752.598333)],
])
def test_wrap_dateline_sinusoidal(pts):
    sinus_crs = geometry.CRS("""PROJCS["unnamed",
                           GEOGCS["Unknown datum based upon the custom spheroid",
                           DATUM["Not specified (based on custom spheroid)", SPHEROID["Custom spheroid",6371007.181,0]],
                           PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],
                           PROJECTION["Sinusoidal"],
                           PARAMETER["longitude_of_center",0],
                           PARAMETER["false_easting",0],
                           PARAMETER["false_northing",0],
                           UNIT["Meter",1]]""")

    wrap = geometry.polygon(pts, crs=sinus_crs)
    wrapped = wrap.to_crs(epsg4326)
    assert wrapped.type == 'Polygon'
    wrapped = wrap.to_crs(epsg4326, wrapdateline=True)
    assert wrapped.type == 'MultiPolygon'
    assert not wrapped.intersects(geometry.line([(0, -90), (0, 90)], crs=epsg4326))


def test_wrap_dateline_utm():
    poly = geometry.box(618300, -1876800,
                        849000, -1642500, 'EPSG:32660')

    wrapped = poly.to_crs(epsg4326)
    assert wrapped.type == 'Polygon'
    assert wrapped.intersects(geometry.line([(0, -90), (0, 90)], crs=epsg4326))
    wrapped = poly.to_crs(epsg4326, wrapdateline=True)
    assert wrapped.type == 'MultiPolygon'
    assert not wrapped.intersects(geometry.line([(0, -90), (0, 90)], crs=epsg4326))


def test_3d_geometry_converted_to_2d_geometry():
    coordinates = [(115.8929714190001, -28.577007674999948, 0.0),
                   (115.90275429200005, -28.57698532699993, 0.0),
                   (115.90412631000004, -28.577577566999935, 0.0),
                   (115.90157040700001, -28.58521105999995, 0.0),
                   (115.89382838900008, -28.585473711999953, 0.0),
                   (115.8929714190001, -28.577007674999948, 0.0)]
    geom_3d = {'coordinates': [coordinates],
               'type': 'Polygon'}
    geom_2d = {'coordinates': [[(x, y) for x, y, z in coordinates]],
               'type': 'Polygon'}

    g_2d = geometry.Geometry(geom_2d)
    g_3d = geometry.Geometry(geom_3d)

    assert {2} == set(len(pt) for pt in g_3d.boundary.coords)  # All coordinates are 2D

    assert g_2d == g_3d  # 3D geometry has been converted to a 2D by dropping the Z axis


def test_3d_point_converted_to_2d_point():
    point = (-35.5029340, 145.9312455, 0.0)

    point_3d = {'coordinates': point,
                'type': 'Point'}
    point_2d = {'coordinates': (point[0], point[1]),
                'type': 'Point'}

    p_2d = geometry.Geometry(point_2d)
    p_3d = geometry.Geometry(point_3d)

    assert len(p_3d.coords[0]) == 2

    assert p_2d == p_3d


def test_crs():
    custom_crs = geometry.CRS("""PROJCS["unnamed",
                           GEOGCS["Unknown datum based upon the custom spheroid",
                           DATUM["Not specified (based on custom spheroid)", SPHEROID["Custom spheroid",6371007.181,0]],
                           PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],
                           PROJECTION["Sinusoidal"],
                           PARAMETER["longitude_of_center",0],
                           PARAMETER["false_easting",0],
                           PARAMETER["false_northing",0],
                           UNIT["Meter",1]]""")

    crs = epsg3577
    assert crs.geographic is False
    assert crs.projected is True
    assert crs.dimensions == ('y', 'x')
    assert crs.epsg == 3577
    assert crs.units == ('metre', 'metre')
    assert isinstance(repr(crs), str)

    crs = epsg4326
    assert crs.geographic is True
    assert crs.projected is False
    assert crs.dimensions == ('latitude', 'longitude')
    assert crs.epsg == 4326

    crs2 = geometry.CRS(crs)
    assert crs2 == crs
    assert crs.proj == crs2.proj

    assert epsg4326.valid_region == geometry.box(-180, -90, 180, 90, epsg4326)
    assert epsg3857.valid_region.crs == epsg4326
    xmin, _, xmax, _ = epsg3857.valid_region.boundingbox
    assert (xmin, xmax) == (-180, 180)
    assert custom_crs.valid_region is None

    assert epsg3577 == epsg3577
    assert epsg3577 == 'EPSG:3577'
    assert (epsg3577 != epsg3577) is False
    assert (epsg3577 == epsg4326) is False
    assert (epsg3577 == 'EPSG:4326') is False
    assert epsg3577 != epsg4326
    assert epsg3577 != 'EPSG:4326'

    bad_crs = ['cupcakes',
               ('PROJCS["unnamed",'
                'GEOGCS["WGS 84", DATUM["WGS_1984", SPHEROID["WGS 84",6378137,298.257223563, AUTHORITY["EPSG","7030"]],'
                'AUTHORITY["EPSG","6326"]], PRIMEM["Greenwich",0, AUTHORITY["EPSG","8901"]],'
                'UNIT["degree",0.0174532925199433, AUTHORITY["EPSG","9122"]], AUTHORITY["EPSG","4326"]]]')]

    for bad in bad_crs:
        with pytest.raises(geometry.CRSError):
            CRS(bad)

    with pytest.warns(DeprecationWarning):
        assert str(epsg3857) == epsg3857.crs_str


def test_polygon_path():
    from datacube.utils.geometry.tools import polygon_path

    pp = polygon_path([0, 1])
    assert pp.shape == (2, 5)
    assert set(pp.ravel()) == {0, 1}

    pp2 = polygon_path([0, 1], [0, 1])
    assert (pp2 == pp).all()

    pp = polygon_path([0, 1], [2, 3])
    assert set(pp[0].ravel()) == {0, 1}
    assert set(pp[1].ravel()) == {2, 3}


def test_gbox_boundary():
    from datacube.utils.geometry.tools import gbox_boundary
    import numpy as np

    xx = np.zeros((2, 6))

    bb = gbox_boundary(xx, 3)

    assert bb.shape == (4 + (3-2)*4, 2)
    assert set(bb.T[0]) == {0.0, 3.0, 6.0}
    assert set(bb.T[1]) == {0.0, 1.0, 2.0}


def test_geobox_scale_down():
    from datacube.utils.geometry import GeoBox, CRS

    crs = CRS('EPSG:3857')

    A = mkA(0, (111.2, 111.2), translation=(125671, 251465))
    for s in [2, 3, 4, 8, 13, 16]:
        gbox = GeoBox(233*s, 755*s, A, crs)
        gbox_ = scaled_down_geobox(gbox, s)

        assert gbox_.width == 233
        assert gbox_.height == 755
        assert gbox_.crs is crs
        assert gbox_.extent.contains(gbox.extent)
        assert gbox.extent.difference(gbox.extent).area == 0.0

    gbox = GeoBox(1, 1, A, crs)
    for s in [2, 3, 5]:
        gbox_ = scaled_down_geobox(gbox, 3)

        assert gbox_.shape == (1, 1)
        assert gbox_.crs is crs
        assert gbox_.extent.contains(gbox.extent)


def test_roi_tools():
    from datacube.utils.geometry import (
        roi_is_empty,
        roi_is_full,
        roi_shape,
        roi_normalise,
        roi_boundary,
        roi_from_points,
        roi_center,
        roi_pad,
        roi_intersect,
        scaled_down_roi,
        scaled_up_roi,
        scaled_down_shape,
    )
    from numpy import s_

    assert roi_shape(s_[2:4, 3:4]) == (2, 1)
    assert roi_shape(s_[:4, :7]) == (4, 7)

    assert roi_is_empty(s_[:4, :5]) is False
    assert roi_is_empty(s_[1:1, :10]) is True
    assert roi_is_empty(s_[7:3, :10]) is True

    assert roi_is_empty(s_[:3]) is False
    assert roi_is_empty(s_[4:4]) is True

    assert roi_is_full(s_[:3], 3) is True
    assert roi_is_full(s_[:3, 0:4], (3, 4)) is True
    assert roi_is_full(s_[:, 0:4], (33, 4)) is True
    assert roi_is_full(s_[1:3, 0:4], (3, 4)) is False
    assert roi_is_full(s_[1:3, 0:4], (2, 4)) is False
    assert roi_is_full(s_[0:4, 0:4], (3, 4)) is False

    roi = s_[0:8, 0:4]
    roi_ = scaled_down_roi(roi, 2)
    assert roi_shape(roi_) == (4, 2)
    assert scaled_down_roi(scaled_up_roi(roi, 3), 3) == roi

    assert scaled_down_shape(roi_shape(roi), 2) == roi_shape(scaled_down_roi(roi, 2))

    assert roi_shape(scaled_up_roi(roi, 10000, (40, 50))) == (40, 50)

    assert roi_normalise(s_[3:4], 40) == s_[3:4]
    assert roi_normalise(s_[:4], (40,)) == s_[0:4]
    assert roi_normalise(s_[:], (40,)) == s_[0:40]
    assert roi_normalise(s_[:-1], (3,)) == s_[0:2]
    assert roi_normalise(s_[-2:-1, :], (10, 20)) == s_[8:9, 0:20]
    assert roi_normalise(s_[-2:-1, :, 3:4], (10, 20, 100)) == s_[8:9, 0:20, 3:4]
    assert roi_center(s_[0:3]) == 1.5
    assert roi_center(s_[0:2, 0:6]) == (1, 3)

    roi = s_[0:2, 4:13]
    xy = roi_boundary(roi)

    assert xy.shape == (4, 2)
    assert roi_from_points(xy, (2, 13)) == roi

    assert roi_intersect(roi, roi) == roi
    assert roi_intersect(s_[0:3], s_[1:7]) == s_[1:3]
    assert roi_intersect(s_[0:3], (s_[1:7],)) == s_[1:3]
    assert roi_intersect((s_[0:3],), s_[1:7]) == (s_[1:3],)

    assert roi_intersect(s_[4:7, 5:6], s_[0:1, 7:8]) == s_[4:4, 6:6]

    assert roi_pad(s_[0:4], 1, 4) == s_[0:4]
    assert roi_pad(s_[0:4, 1:5], 1, (4, 6)) == s_[0:4, 0:6]
    assert roi_pad(s_[2:3, 1:5], 10, (7, 9)) == s_[0:7, 0:9]


def test_apply_affine():
    A = mkA(rot=10, scale=(3, 1.3), translation=(-100, +2.3))
    xx, yy = np.meshgrid(np.arange(13), np.arange(11))

    xx_, yy_ = apply_affine(A, xx, yy)

    assert xx_.shape == xx.shape
    assert yy_.shape == xx.shape

    xy_expect = [A*(x, y) for x, y in zip(xx.ravel(), yy.ravel())]
    xy_got = [(x, y) for x, y in zip(xx_.ravel(), yy_.ravel())]

    np.testing.assert_array_almost_equal(xy_expect, xy_got)


def test_point_transformer():
    from datacube.utils.geometry import point

    tr = epsg3857.transformer_to_crs(epsg4326)
    tr_back = epsg4326.transformer_to_crs(epsg3857)

    pts = [(0, 0), (0, 1),
           (1, 2), (10, 11)]
    x, y = np.vstack(pts).astype('float64').T

    pts_expect = [point(*pt, epsg3857).to_crs(epsg4326).points[0]
                  for pt in pts]

    x_expect = [pt[0] for pt in pts_expect]
    y_expect = [pt[1] for pt in pts_expect]

    x_, y_ = tr(x, y)
    assert x_.shape == x.shape
    np.testing.assert_array_almost_equal(x_, x_expect)
    np.testing.assert_array_almost_equal(y_, y_expect)

    x, y = (a.reshape(2, 2) for a in (x, y))
    x_, y_ = tr(x, y)
    assert x_.shape == x.shape

    xb, yb = tr_back(x_, y_)
    np.testing.assert_array_almost_equal(x, xb)
    np.testing.assert_array_almost_equal(y, yb)

    # check nans
    x_, y_ = tr(np.asarray([np.nan, 0, np.nan]),
                np.asarray([0, np.nan, np.nan]))

    assert np.isnan(x_).all()
    assert np.isnan(y_).all()


def test_split_translation():

    def verify(a, b):
        a = np.asarray(a)
        b = np.asarray(b)
        np.testing.assert_array_almost_equal(a, b)

    def tt(tx, ty, *expect):
        verify(split_translation((tx, ty)), expect)

    assert split_translation((1, 2)) == ((1, 2), (0, 0))
    assert split_translation((-1, -2)) == ((-1, -2), (0, 0))
    tt(1.3, 2.5, (1, 2), (0.3, 0.5))
    tt(1.1, 2.6, (1, 3), (0.1, -0.4))
    tt(-1.1, 2.8, (-1, 3), (-0.1, -0.2))
    tt(-1.9, 2.05, (-2, 2), (+0.1, 0.05))
    tt(-1.5, 2.45, (-1, 2), (-0.5, 0.45))


def get_diff(A, B):
    from math import sqrt
    return sqrt(sum((a-b)**2 for a, b in zip(A, B)))


def test_affine_checks():
    assert is_affine_st(mkA(scale=(1, 2), translation=(3, -10))) is True
    assert is_affine_st(mkA(scale=(1, -2), translation=(-3, -10))) is True
    assert is_affine_st(mkA(rot=0.1)) is False
    assert is_affine_st(mkA(shear=0.4)) is False


def test_affine_rsw():

    def run_test(a, scale, shear=0, translation=(0, 0), tol=1e-8):
        A = mkA(a, scale=scale, shear=shear, translation=translation)

        R, W, S = decompose_rws(A)

        assert get_diff(A, R*W*S) < tol
        assert get_diff(S, mkA(0, scale)) < tol
        assert get_diff(R, mkA(a, translation=translation)) < tol

    for a in (0, 12, 45, 33, 67, 89, 90, 120, 170):
        run_test(a, (1, 1))
        run_test(a, (0.5, 2))
        run_test(-a, (0.5, 2))

        run_test(a, (1, 2))
        run_test(-a, (1, 2))

        run_test(a, (2, -1))
        run_test(-a, (2, -1))

    run_test(0, (3, 4), 10)
    run_test(-33, (3, -1), 10, translation=(100, -333))


def test_fit():
    from random import uniform

    def run_test(A, n, tol=1e-5):
        X = [(uniform(0, 1), uniform(0, 1))
             for _ in range(n)]
        Y = [A*x for x in X]  # noqa: N806
        A_ = affine_from_pts(X, Y)  # noqa: N806

        assert get_diff(A, A_) < tol

    A = mkA(13, scale=(3, 4), shear=3, translation=(100, -3000))

    run_test(A, 3)
    run_test(A, 10)

    run_test(mkA(), 3)
    run_test(mkA(), 10)


def test_scale_at_point():
    def mk_transform(sx, sy):
        A = mkA(37, scale=(sx, sy), translation=(2127, 93891))

        def transofrom(pts):
            return [A*x for x in pts]

        return transofrom

    tol = 1e-4
    pt = (0, 0)
    for sx, sy in [(3, 4), (0.4, 0.333)]:
        tr = mk_transform(sx, sy)
        sx_, sy_ = get_scale_at_point(pt, tr)
        assert abs(sx - sx_) < tol
        assert abs(sy - sy_) < tol

        sx_, sy_ = get_scale_at_point(pt, tr, 0.1)
        assert abs(sx - sx_) < tol
        assert abs(sy - sy_) < tol


def test_pix_transform():
    pt = tuple([int(x/10)*10 for x in
                geometry.point(145, -35, epsg4326).to_crs(epsg3577).coords[0]])

    A = mkA(scale=(20, -20), translation=pt)

    src = geometry.GeoBox(1024, 512, A, epsg3577)
    dst = geometry.GeoBox.from_geopolygon(src.geographic_extent,
                                          (0.0001, -0.0001))

    tr = native_pix_transform(src, dst)

    pts_src = [(0, 0), (10, 20), (300, 200)]
    pts_dst = tr(pts_src)
    pts_src_ = tr.back(pts_dst)

    np.testing.assert_almost_equal(pts_src, pts_src_)
    assert tr.linear is None

    # check identity transform
    tr = native_pix_transform(src, src)

    pts_src = [(0, 0), (10, 20), (300, 200)]
    pts_dst = tr(pts_src)
    pts_src_ = tr.back(pts_dst)

    np.testing.assert_almost_equal(pts_src, pts_src_)
    np.testing.assert_almost_equal(pts_src, pts_dst)
    assert tr.linear is not None
    assert tr.back.linear is not None
    assert tr.back.back is tr

    # check scale only change
    tr = native_pix_transform(src, scaled_down_geobox(src, 2))
    pts_dst = tr(pts_src)
    pts_src_ = tr.back(pts_dst)

    assert tr.linear is not None
    assert tr.back.linear is not None
    assert tr.back.back is tr

    np.testing.assert_almost_equal(pts_dst,
                                   [(x/2, y/2) for (x, y) in pts_src])

    np.testing.assert_almost_equal(pts_src, pts_src_)


def test_compute_reproject_roi():
    src = AlbersGS.tile_geobox((15, -40))
    dst = geometry.GeoBox.from_geopolygon(src.extent.to_crs(epsg3857).buffer(10),
                                          resolution=src.resolution)

    rr = compute_reproject_roi(src, dst)

    assert rr.roi_src == np.s_[0:src.height, 0:src.width]
    assert 0 < rr.scale < 1
    assert rr.is_st is False
    assert rr.transform.linear is None
    assert rr.scale in rr.scale2

    # check pure translation case
    roi_ = np.s_[113:-100, 33:-10]
    rr = compute_reproject_roi(src, src[roi_])
    assert rr.roi_src == roi_normalise(roi_, src.shape)
    assert rr.scale == 1
    assert rr.is_st is True

    rr = compute_reproject_roi(src, src[roi_], padding=0, align=0)
    assert rr.roi_src == roi_normalise(roi_, src.shape)
    assert rr.scale == 1
    assert rr.scale2 == (1, 1)

    # check pure translation case
    roi_ = np.s_[113:-100, 33:-10]
    rr = compute_reproject_roi(src, src[roi_], align=256)

    assert rr.roi_src == np.s_[0:src.height, 0:src.width]
    assert rr.scale == 1

    roi_ = np.s_[113:-100, 33:-10]
    rr = compute_reproject_roi(src, src[roi_])

    assert rr.scale == 1
    assert roi_shape(rr.roi_src) == roi_shape(rr.roi_dst)
    assert roi_shape(rr.roi_dst) == src[roi_].shape


def test_compute_reproject_roi_issue647():
    """ In some scenarios non-overlapping geoboxes will result in non-empty
    `roi_dst` even though `roi_src` is empty.

    Test this case separately.
    """
    from datacube.utils.geometry import CRS

    src = GeoBox(10980, 10980,
                 Affine(10, 0, 300000,
                        0, -10, 5900020),
                 CRS('epsg:32756'))

    dst = GeoBox(976, 976, Affine(10, 0, 1730240,
                                  0, -10, -4170240),
                 CRS('EPSG:3577'))

    assert src.extent.overlaps(dst.extent.to_crs(src.crs)) is False

    rr = compute_reproject_roi(src, dst)

    assert roi_is_empty(rr.roi_src)
    assert roi_is_empty(rr.roi_dst)


def test_compute_reproject_roi_issue1047():
    """ `compute_reproject_roi(geobox, geobox[roi])` sometimes returns
    `src_roi != roi`, when `geobox` has (1) tiny pixels and (2) oddly
    sized `alignment`.

    Test this issue is resolved.
    """
    geobox = GeoBox(3000, 3000,
                    Affine(0.00027778, 0.0, 148.72673054908861,
                           0.0, -0.00027778, -34.98825802556622), "EPSG:4326")
    src_roi = np.s_[2800:2810, 10:30]
    rr = compute_reproject_roi(geobox, geobox[src_roi])

    assert rr.is_st is True
    assert rr.roi_src == src_roi
    assert rr.roi_dst == np.s_[0:10, 0:20]


def test_window_from_slice():
    from numpy import s_

    assert w_[None] is None
    assert w_[s_[:3, 4:5]] == ((0, 3), (4, 5))
    assert w_[s_[0:3, :5]] == ((0, 3), (0, 5))
    assert w_[list(s_[0:3, :5])] == ((0, 3), (0, 5))

    for roi in [s_[:3], s_[:3, :4, :5], 0]:
        with pytest.raises(ValueError):
            w_[roi]


def test_axis_overlap():
    s_ = np.s_

    # Source overlaps destination fully
    #
    # S: |<--------------->|
    # D:      |<----->|
    assert compute_axis_overlap(100, 20, 1, 10) == s_[10:30, 0:20]
    assert compute_axis_overlap(100, 20, 2, 10) == s_[10:50, 0:20]
    assert compute_axis_overlap(100, 20, 0.25, 10) == s_[10:15, 0:20]
    assert compute_axis_overlap(100, 20, -1, 80) == s_[60:80, 0:20]
    assert compute_axis_overlap(100, 20, -0.5, 50) == s_[40:50, 0:20]
    assert compute_axis_overlap(100, 20, -2, 90) == s_[50:90, 0:20]

    # Destination overlaps source fully
    #
    # S:      |<-------->|
    # D: |<----------------->|
    assert compute_axis_overlap(10, 100, 1, -10) == s_[0:10, 10:20]
    assert compute_axis_overlap(10, 100, 2, -10) == s_[0:10, 5:10]
    assert compute_axis_overlap(10, 100, 0.5, -10) == s_[0:10, 20:40]
    assert compute_axis_overlap(10, 100, -1, 11) == s_[0:10, 1:11]

    # Partial overlaps
    #
    # S: |<----------->|
    # D:     |<----------->|
    assert compute_axis_overlap(10, 10, 1, 3) == s_[3:10, 0:7]
    assert compute_axis_overlap(10, 15, 1, 3) == s_[3:10, 0:7]

    # S:     |<----------->|
    # D: |<----------->|
    assert compute_axis_overlap(10, 10, 1, -5) == s_[0:5, 5:10]
    assert compute_axis_overlap(50, 10, 1, -5) == s_[0:5, 5:10]

    # No overlaps
    # S: |<--->|
    # D:         |<--->|
    assert compute_axis_overlap(10, 10, 1, 11) == s_[10:10, 0:0]
    assert compute_axis_overlap(10, 40, 1, 11) == s_[10:10, 0:0]

    # S:         |<--->|
    # D: |<--->|
    assert compute_axis_overlap(10, 10, 1, -11) == s_[0:0, 10:10]
    assert compute_axis_overlap(40, 10, 1, -11) == s_[0:0, 10:10]


def test_crs_compat():
    import rasterio.crs

    crs = CRS("epsg:3577")
    assert crs.epsg == 3577
    crs2 = CRS(crs)
    assert crs.epsg == crs2.epsg

    crs_rio = rasterio.crs.CRS(init='epsg:3577')
    assert CRS(crs_rio).epsg == 3577

    assert (CRS(crs_rio) == crs_rio) is True

    with pytest.raises(geometry.CRSError):
        CRS(("random", "tuple"))

    crs = CRS("epsg:3857")
    with pytest.warns(UserWarning):
        crs_dict = crs.proj.to_dict()

    assert CRS(crs_dict) == crs


def test_crs_hash():
    crs = CRS("epsg:3577")
    crs2 = CRS(crs)

    assert crs is not crs2
    assert len(set([crs, crs2])) == 1


def test_base_internals():
    assert _make_crs_key("epsg:3577") == "EPSG:3577"
    no_epsg_crs = CRS(SAMPLE_WKT_WITHOUT_AUTHORITY)
    assert _make_crs_key(no_epsg_crs.proj) == no_epsg_crs.proj.to_wkt()

    gjson_bad = {'type': 'a', 'coordinates': [1, [2, 3, 4]]}
    assert force_2d(gjson_bad) == {'type': 'a', 'coordinates': [1, [2, 3]]}

    with pytest.raises(ValueError):
        force_2d({'type': 'a', 'coordinates': [set("not a valid element")]})

    assert _round_to_res(0.2, 1.0) == 1
    assert _round_to_res(0.0, 1.0) == 0
    assert _round_to_res(0.05, 1.0) == 0

    assert _norm_crs(None) is None

    with pytest.raises(ValueError):
        _norm_crs_or_error(None)


def test_geom_clone():
    b = geometry.box(0, 0, 10, 20, epsg4326)
    assert b == b.clone()
    assert b.geom is not b.clone().geom

    assert b == geometry.Geometry(b)
    assert b.geom is not geometry.Geometry(b).geom


def test_crs_units_per_degree():
    assert crs_units_per_degree('EPSG:3857', (0, 0)) == crs_units_per_degree('EPSG:3857', 0, 0)
    assert crs_units_per_degree('EPSG:4326', (120, -10)) == approx(1.0, 1e-6)

    assert crs_units_per_degree('EPSG:3857', 0, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree('EPSG:3857', 20, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree('EPSG:3857', 30, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree('EPSG:3857', 180, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree('EPSG:3857', -180, 0) == approx(111319.49, 0.5)


def test_rio_crs__no_epsg():
    import rasterio.crs
    rio_crs = rasterio.crs.CRS.from_wkt(
        'PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",'
        'DATUM["Not specified (based on custom spheroid)",'
        'SPHEROID["Custom spheroid",6371007.181,0]],'
        'PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],'
        'PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],'
        'PARAMETER["false_easting",0],PARAMETER["false_northing",0],'
        'UNIT["Meter",1],AXIS["Easting",EAST],AXIS["Northing",NORTH]]'
    )
    assert CRS(rio_crs).epsg is None


@pytest.mark.parametrize("left, right, off, res, expect", [
    (20, 30, 10, 0, (20, 1)),
    (20, 30.5, 10, 0, (20, 1)),
    (20, 31.5, 10, 0, (20, 2)),
    (20, 30, 10, 3, (13, 2)),
    (20, 30, 10, -3, (17, 2)),
    (20, 30, -10, 0, (30, 1)),
    (19.5, 30, -10, 0, (30, 1)),
    (18.5, 30, -10, 0, (30, 2)),
    (20, 30, -10, 3, (33, 2)),
    (20, 30, -10, -3, (37, 2)),
])
def test_align_pix(left, right, off, res, expect):
    assert _align_pix(left, right, off, res) == expect


def test_lonlat_bounds():
    # example from landsat scene: spans lon=180
    poly = geometry.box(618300, -1876800, 849000, -1642500, 'EPSG:32660')

    bb = geometry.lonlat_bounds(poly)
    assert bb.left < 180 < bb.right
    assert geometry.lonlat_bounds(poly) == geometry.lonlat_bounds(poly, resolution=1e+8)

    bb = geometry.lonlat_bounds(poly, mode='quick')
    assert bb.right - bb.left > 180

    poly = geometry.box(1, -10, 2, 20, 'EPSG:4326')
    assert geometry.lonlat_bounds(poly) == poly.boundingbox

    with pytest.raises(ValueError):
        geometry.lonlat_bounds(geometry.box(0, 0, 1, 1, None))

    multi = {
        "type": "MultiPolygon",
        "coordinates": [
            [[[174, 52], [174, 53], [175, 53], [174, 52]]],
            [[[168, 54], [167, 55], [167, 54], [168, 54]]]
        ]
    }

    multi_geom = geometry.Geometry(multi, "epsg:4326")
    multi_geom_projected = multi_geom.to_crs('epsg:32659', math.inf)

    ll_bounds = geometry.lonlat_bounds(multi_geom)
    ll_bounds_projected = geometry.lonlat_bounds(multi_geom_projected)

    assert ll_bounds == approx(ll_bounds_projected)


@pytest.mark.xfail(True, reason="Bounds computation for large geometries in safe mode is broken")
def test_lonalt_bounds_more_than_180():
    poly = geometry.box(-150, -30, 150, 30, epsg4326).to_crs(epsg3857, math.inf)

    assert geometry.lonlat_bounds(poly, "quick") == approx((-150, -30, 150, 30))
    assert geometry.lonlat_bounds(poly, "safe") == approx((-150, -30, 150, 30))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""
import pytest
import configparser
from textwrap import dedent

from datacube.config import LocalConfig, parse_connect_url, parse_env_params, auto_config
from datacube.testutils import write_files


def test_find_config():
    files = write_files({
        'base.conf': dedent("""\
            [datacube]
            db_hostname: fakehost.test.lan
        """),
        'override.conf': dedent("""\
            [datacube]
            db_hostname: overridden.test.lan
            db_database: overridden_db
        """)
    })

    # One config file
    config = LocalConfig.find(paths=[str(files.joinpath('base.conf'))])
    assert config['db_hostname'] == 'fakehost.test.lan'
    # Not set: uses default
    assert config['db_database'] == 'datacube'

    # Now two config files, with the latter overriding earlier options.
    config = LocalConfig.find(paths=[str(files.joinpath('base.conf')),
                                     str(files.joinpath('override.conf'))])
    assert config['db_hostname'] == 'overridden.test.lan'
    assert config['db_database'] == 'overridden_db'


config_sample = """
[default]
db_database: datacube

# A blank host will use a local socket. Specify a hostname (such as localhost) to use TCP.
db_hostname:

# Credentials are optional: you might have other Postgres authentication configured.
# The default username is the current user id
# db_username:
# A blank password will fall back to default postgres driver authentication, such as reading your ~/.pgpass file.
# db_password:
index_driver: pg


## Development environment ##
[dev]
# These fields are all the defaults, so they could be omitted, but are here for reference

db_database: datacube

# A blank host will use a local socket. Specify a hostname (such as localhost) to use TCP.
db_hostname:

# Credentials are optional: you might have other Postgres authentication configured.
# The default username is the current user id
# db_username:
# A blank password will fall back to default postgres driver authentication, such as reading your ~/.pgpass file.
# db_password:

## Staging environment ##
[staging]
db_hostname: staging.dea.ga.gov.au
"""


def test_using_configparser(tmpdir):
    sample_config = tmpdir.join('datacube.conf')
    sample_config.write(config_sample)

    config = configparser.ConfigParser()
    config.read(str(sample_config))


def test_empty_configfile(tmpdir):
    default_only = """[default]"""
    sample_file = tmpdir.join('datacube.conf')
    sample_file.write(default_only)
    config = configparser.ConfigParser()
    config.read(str(sample_file))


def test_parse_db_url():

    assert parse_connect_url('postgresql:///db') == dict(database='db', hostname='')
    assert parse_connect_url('postgresql://some.tld/db') == dict(database='db', hostname='some.tld')
    assert parse_connect_url('postgresql://some.tld:3344/db') == dict(
        database='db',
        hostname='some.tld',
        port='3344')
    assert parse_connect_url('postgresql://user@some.tld:3344/db') == dict(
        username='user',
        database='db',
        hostname='some.tld',
        port='3344')
    assert parse_connect_url('postgresql://user:pass@some.tld:3344/db') == dict(
        password='pass',
        username='user',
        database='db',
        hostname='some.tld',
        port='3344')

    # check urlencode is reversed for password field
    assert parse_connect_url('postgresql://user:pass%40@some.tld:3344/db') == dict(
        password='pass@',
        username='user',
        database='db',
        hostname='some.tld',
        port='3344')

    assert parse_connect_url('postgresql:///db?host=/var/run/postgresql') == dict(
        database='db',
        hostname='/var/run/postgresql')

    assert parse_connect_url(
        'postgresql:///?user=user&password=pass%40&host=/var/run/postgresql&port=3344&dbname=db&sslmode=allow'
    ) == dict(
        password='pass@',
        username='user',
        database='db',
        hostname='/var/run/postgresql',
        port='3344')


def _clear_cfg_env(monkeypatch):
    for e in ('DATACUBE_DB_URL',
              'DB_HOSTNAME',
              'DB_PORT',
              'DB_USERNAME',
              'DB_PASSWORD',
              'DATACUBE_IAM_AUTHENTICATION',
              'DATACUBE_IAM_TIMEOUT'):
        monkeypatch.delenv(e, raising=False)


def test_parse_env(monkeypatch):
    def set_env(**kw):
        _clear_cfg_env(monkeypatch)
        for e, v in kw.items():
            monkeypatch.setenv(e, v)

    def check_env(**kw):
        set_env(**kw)
        return parse_env_params()

    assert check_env() == {}
    assert check_env(DATACUBE_IAM_AUTHENTICATION="yes",
                     DATACUBE_IAM_TIMEOUT='666') == dict(
        iam_authentication=True,
        iam_timeout=666)
    assert check_env(DATACUBE_DB_URL='postgresql:///db') == dict(
        hostname='',
        database='db'
    )
    assert check_env(DATACUBE_DB_URL='postgresql://uu:%20pass%40@host.tld:3344/db') == dict(
        username='uu',
        password=' pass@',
        hostname='host.tld',
        port='3344',
        database='db'
    )
    assert check_env(DB_DATABASE='db') == dict(
        database='db'
    )
    assert check_env(DB_DATABASE='db', DB_HOSTNAME='host.tld') == dict(
        database='db',
        hostname='host.tld'
    )
    assert check_env(DB_DATABASE='db',
                     DB_HOSTNAME='host.tld',
                     DB_USERNAME='user',
                     DB_PASSWORD='pass@') == dict(
                         database='db',
                         hostname='host.tld',
                         username='user',
                         password='pass@')

    assert check_env(DB_DATABASE='db',
                     DB_HOSTNAME='host.tld',
                     DB_USERNAME='user',
                     DB_PORT='',
                     DB_PASSWORD='pass@') == dict(
                         database='db',
                         hostname='host.tld',
                         username='user',
                         password='pass@')


def test_cfg_from_env(monkeypatch):
    def set_env(**kw):
        _clear_cfg_env(monkeypatch)
        for e, v in kw.items():
            monkeypatch.setenv(e, v)

    set_env(DATACUBE_DB_URL='postgresql://uu:%20pass%40@host.tld:3344/db')
    cfg = LocalConfig.find()
    assert '3344' in str(cfg)
    assert '3344' in repr(cfg)
    assert cfg['db_username'] == 'uu'
    assert cfg['db_password'] == ' pass@'
    assert cfg['db_hostname'] == 'host.tld'
    assert cfg['db_database'] == 'db'
    assert cfg['db_port'] == '3344'

    # check that password is redacted
    assert " pass@" not in str(cfg)
    assert "***" in str(cfg)
    assert " pass@" not in repr(cfg)
    assert "***" in repr(cfg)

    set_env(DB_DATABASE='dc2',
            DB_HOSTNAME='remote.db',
            DB_PORT='4433',
            DB_USERNAME='dcu',
            DB_PASSWORD='gg')
    cfg = LocalConfig.find()
    assert cfg['db_username'] == 'dcu'
    assert cfg['db_password'] == 'gg'
    assert cfg['db_hostname'] == 'remote.db'
    assert cfg['db_database'] == 'dc2'
    assert cfg['db_port'] == '4433'


def test_auto_config(monkeypatch, tmpdir):
    from pathlib import Path

    cfg_file = Path(str(tmpdir/"dc.cfg"))
    assert cfg_file.exists() is False
    cfg_file_name = str(cfg_file)

    _clear_cfg_env(monkeypatch)
    monkeypatch.setenv('DATACUBE_CONFIG_PATH', cfg_file_name)

    assert auto_config() == cfg_file_name
    assert cfg_file.exists() is True

    monkeypatch.setenv('DB_HOSTNAME', 'should-not-be-used.local')
    # second run should skip overwriting
    assert auto_config() == cfg_file_name

    config = LocalConfig.find(paths=cfg_file_name)
    assert config['db_hostname'] == ''
    assert config['db_database'] == 'datacube'

    cfg_file.unlink()
    assert cfg_file.exists() is False
    _clear_cfg_env(monkeypatch)

    monkeypatch.setenv('DATACUBE_CONFIG_PATH', cfg_file_name)
    monkeypatch.setenv('DB_HOSTNAME', 'some.db')
    monkeypatch.setenv('DB_USERNAME', 'user')

    assert auto_config() == cfg_file_name
    config = LocalConfig.find(paths=cfg_file_name)
    assert config['db_hostname'] == 'some.db'
    assert config['db_database'] == 'datacube'
    assert config['db_username'] == 'user'

    assert config.get('no_such_key', 10) == 10
    with pytest.raises(configparser.NoOptionError):
        config.get('no_such_key')



import pytest

from unittest.mock import MagicMock

from datacube.index.hl import Doc2Dataset


def test_support_validation(non_geo_dataset_doc, eo_dataset_doc):
    idx = MagicMock()

    idx.supports_legacy = False
    idx.supports_nongeo = False
    with pytest.raises(ValueError, match="EO3 cannot be set to False"):
        resolver = Doc2Dataset(idx, eo3=False)

    idx.supports_legacy = True
    idx.supports_nongeo = False
    resolver = Doc2Dataset(idx, products=["product_a"], eo3=False)
    _, err = resolver(non_geo_dataset_doc, "//location/")
    assert "Non-geospatial metadata formats" in err

    idx.supports_legacy = False
    idx.supports_nongeo = True
    resolver = Doc2Dataset(idx, products=["product_a"], eo3=False)
    _, err = resolver(eo_dataset_doc, "//location/")
    assert "Legacy metadata formats" in err



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

from datacube.ui.task_app import task_app, run_tasks, wrap_task
import datacube.executor


def make_test_config(index, config, **kwargs):
    assert index == 'Fake Index'
    assert 'config_arg' in kwargs

    config['some_item'] = 'make_test_config'
    config['num_tasks'] = config.get('num_tasks', 3)
    return config


def make_test_tasks(index, config, **kwargs):
    assert index == 'Fake Index'
    assert 'task_arg' in kwargs

    num_tasks = config['num_tasks']
    for i in range(0, num_tasks):
        yield 'Task: {}'.format(i)


@task_app(make_config=make_test_config, make_tasks=make_test_tasks)
def my_test_app(index, config, tasks, **kwargs):
    assert index == 'Fake Index'
    assert config['some_item'] == 'make_test_config'
    assert 'app_arg' in kwargs

    task_list = list(tasks)
    assert len(task_list) == config['num_tasks']


def test_task_app(tmpdir):
    index = 'Fake Index'

    app_config = tmpdir.join("app_config.yaml")
    app_config.write('name: Test Config\r\n'
                     'description: This is my test app config file')

    my_test_app(index, str(app_config), app_arg=True, config_arg=True, task_arg=True)


def test_task_app_with_task_file(tmpdir):
    index = 'Fake Index'

    app_config = tmpdir.join("app_config.yaml")
    app_config.write('name: Test Config\r\n'
                     'description: This is my test app config file')

    taskfile = tmpdir.join("tasks.bin")
    assert not taskfile.check()

    my_test_app(index, app_config=str(app_config), output_tasks_file=str(taskfile), config_arg=True, task_arg=True)

    assert taskfile.check()

    my_test_app(index, input_tasks_file=str(taskfile), app_arg=True)


def test_task_app_with_no_tasks(tmpdir):
    index = 'Fake Index'

    app_config = tmpdir.join("app_config.yaml")
    app_config.write('name: Test Config\r\n'
                     'description: This is my test app config file\r\n'
                     'num_tasks: 0')

    taskfile = tmpdir.join("tasks.bin")
    assert not taskfile.check()

    my_test_app(index, app_config=str(app_config), output_tasks_file=str(taskfile), config_arg=True, task_arg=True)

    assert not taskfile.check()


def test_task_app_year_splitting():
    import pandas as pd
    from datacube.ui.task_app import validate_year, break_query_into_years
    one_millisecond = pd.Timedelta('1 ms')

    def is_close(ts1, ts2, max_delta=one_millisecond):
        return abs(pd.Timestamp(ts1) - pd.Timestamp(ts2)) < max_delta

    assert validate_year(None, None, None) is None
    year_range = validate_year(None, None, '1996-2004')
    assert is_close(year_range[0], '1996-01-01 00:00:00')
    assert is_close(year_range[1], '2004-12-31 23:59:59.999999999')

    year_range = validate_year(None, None, '2003')
    assert is_close(year_range[0], '2003-01-01 00:00:00')
    assert is_close(year_range[1], '2003-12-31 23:59:59.999999999')

    # Test that a no year range makes a single query
    year_range = None
    query = break_query_into_years(year_range)
    assert len(query) == 1
    assert query[0] == {}

    # Test that a no year range makes a single query with additional params
    year_range = None
    test_cell_index = (11, 12)
    query = break_query_into_years(year_range, cell_index=test_cell_index)
    assert len(query) == 1
    assert query[0] == {'cell_index': test_cell_index}

    # Test that a single year makes a single query
    year_range = ("1996", "1996")
    query = break_query_into_years(year_range)
    assert len(query) == 1
    assert is_close(query[0]['time'][0], '1996-01-01 00:00:00')
    assert is_close(query[0]['time'][1], '1996-12-31 23:59:59.999999999')

    # Test that a multiple years makes multiple queries
    year_range = ("1996", "1997")
    query = break_query_into_years(year_range)
    assert len(query) == 2
    assert is_close(query[0]['time'][0], '1996-01-01 00:00:00')
    assert is_close(query[0]['time'][1], '1996-12-31 23:59:59.999999999')
    assert is_close(query[1]['time'][0], '1997-01-01 00:00:00')
    assert is_close(query[1]['time'][1], '1997-12-31 23:59:59.999999999')

    # Check that additional kwargs can be used in the query
    year_range = ("1996", "1997")
    test_cell_index = (11, 12)
    query = break_query_into_years(year_range, cell_index=test_cell_index)
    assert len(query) == 2
    assert is_close(query[0]['time'][0], '1996-01-01 00:00:00')
    assert is_close(query[0]['time'][1], '1996-12-31 23:59:59.999999999')
    assert query[0]['cell_index'] == test_cell_index
    assert is_close(query[1]['time'][0], '1997-01-01 00:00:00')
    assert is_close(query[1]['time'][1], '1997-12-31 23:59:59.999999999')
    assert query[1]['cell_index'] == test_cell_index


def test_task_app_cell_index(tmpdir):
    from datacube.ui.task_app import validate_cell_index, validate_cell_list, cell_list_to_file

    assert validate_cell_index(None, None, None) is None
    assert validate_cell_index(None, None, '17,-12') == (17, -12)

    cell_list = [(17, 12), (16, -10), (-23, 0)]

    cell_list_file = tmpdir.join('cell_list.txt')
    if cell_list_file.exists():
        cell_list_file.unlink()
    assert not cell_list_file.check()

    cell_list_to_file(str(cell_list_file), cell_list)

    assert cell_list_file.check()

    assert validate_cell_list(None, None, None) is None
    assert validate_cell_list(None, None, str(cell_list_file)) == cell_list


def test_run_tasks():
    executor = datacube.executor.SerialExecutor()
    tasks = ({'val': i} for i in range(3))
    tasks_to_do = list(range(3))

    def task_func(task):
        x = task['val']
        return (x, x**2)

    def process_result_func(result):
        assert result[0]**2 == result[1]
        tasks_to_do.remove(result[0])

    run_tasks(tasks, executor, task_func, process_result_func)
    assert not tasks_to_do

    # no task proc specified
    tasks = ({'val': i} for i in range(3))
    run_tasks(tasks, executor, task_func)


def test_wrap_task():
    def task_with_args(task, a, b):
        return (task, a, b)

    assert task_with_args(1, 2, 'a') == (1, 2, 'a')
    assert wrap_task(task_with_args, 'a', 'b')(0) == (0, 'a', 'b')



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
""" Defines abstract types for IO drivers.
"""
from typing import (
    List, Tuple, Optional, Union, Any, Iterable,
    TYPE_CHECKING
)

from abc import ABCMeta, abstractmethod
import numpy as np
from affine import Affine
from concurrent.futures import Future
from datacube.storage import BandInfo
from datacube.utils.geometry import CRS

# pylint: disable=invalid-name,unsubscriptable-object,pointless-statement

if TYPE_CHECKING:
    FutureGeoRasterReader = Future['GeoRasterReader']  # pragma: no cover
    FutureNdarray = Future[np.ndarray]                 # pragma: no cover
else:
    FutureGeoRasterReader = Future
    FutureNdarray = Future


RasterShape = Tuple[int, int]
RasterWindow = Tuple[slice, slice]


class GeoRasterReader(object, metaclass=ABCMeta):
    """ Abstract base class for dataset reader.
    """

    @property
    @abstractmethod
    def crs(self) -> Optional[CRS]:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def transform(self) -> Optional[Affine]:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def shape(self) -> RasterShape:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def nodata(self) -> Optional[Union[int, float]]:
        ...  # pragma: no cover

    @abstractmethod
    def read(self,
             window: Optional[RasterWindow] = None,
             out_shape: Optional[RasterShape] = None) -> FutureNdarray:
        ...  # pragma: no cover


class ReaderDriver(object, metaclass=ABCMeta):
    """ Interface for Reader Driver
    """

    @abstractmethod
    def new_load_context(self,
                         bands: Iterable[BandInfo],
                         old_ctx: Optional[Any]) -> Any:
        """Recycle old context if available/possible and create new context.
           ``old_ctx`` won't be used after this call.

           Same context object is passed to all calls to ``open`` function that
           happen within the same ``dc.load``.

           If your driver doesn't need it just return ``None``
        """
        ...  # pragma: no cover

    @abstractmethod
    def open(self, band: BandInfo, ctx: Any) -> FutureGeoRasterReader:
        ...  # pragma: no cover


class ReaderDriverEntry(object, metaclass=ABCMeta):
    @property
    @abstractmethod
    def protocols(self) -> List[str]:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def formats(self) -> List[str]:
        ...  # pragma: no cover

    @abstractmethod
    def supports(self, protocol: str, fmt: str) -> bool:
        ...  # pragma: no cover

    def new_instance(self, cfg: dict) -> ReaderDriver:
        ...  # pragma: no cover



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from collections import OrderedDict
from pathlib import Path

import pytest
from datacube.model import DatasetType
from datacube.utils.documents import read_documents

PROJECT_ROOT = Path(__file__).parents[1]
GEDI_PRODUCT = (
    PROJECT_ROOT / "tests" / "data" / "lbg" / "gedi" / "GEDI02_B_3d_format.yaml"
)
# GEDI dataset type yaml path


@pytest.fixture
def cover_z_dataset_type():
    """The GEDI dataset type for cover_z."""
    for doc in read_documents(GEDI_PRODUCT):
        if doc[1]["name"] == "gedi_l2b_cover_z":
            yield doc[1]
            break
    # pytest will raise an error if nothing was yielded


def test_extra_dimensions(eo3_metadata, cover_z_dataset_type):
    """Check the ExtraDimensions class."""
    dt = DatasetType(eo3_metadata, cover_z_dataset_type)

    # Check dims
    dt.extra_dimensions.dims == OrderedDict(
        [("z", {"name": "z", "values": list(range(5, 151, 5)), "dtype": "float64"})]
    )

    # Check original slice
    assert dt.extra_dimensions.dim_slice == {"z": (0, 30)}

    # Check measurements values
    assert dt.extra_dimensions.measurements_values("z") == list(range(5, 151, 5))

    # User-selected slicing
    sliced = dt.extra_dimensions[{"z": (5, 12)}]
    assert sliced.dims == OrderedDict(
        [("z", {"name": "z", "values": [5, 10], "dtype": "float64"})]
    )

    # Check measurements_slice
    assert dt.extra_dimensions.measurements_slice("z") == slice(0, 30, None)

    # Check measurements index
    assert dt.extra_dimensions.measurements_index("z") == (0, 30)

    # Check index_of
    assert dt.extra_dimensions.index_of("z", 50) == 9

    # Check coord slice
    assert dt.extra_dimensions.coord_slice("z", 50) == (9, 10)
    assert dt.extra_dimensions.coord_slice("z", 48.3) == (9, 9)
    assert dt.extra_dimensions.coord_slice("z", (48.3, 62)) == (9, 12)
    assert dt.extra_dimensions.coord_slice("z", (148.3, 162)) == (29, 30)
    assert dt.extra_dimensions.coord_slice("z", 1000) == (30, 30)
    assert dt.extra_dimensions.coord_slice("z", (1000, 2000)) == (30, 30)

    # Check chunk size
    assert dt.extra_dimensions.chunk_size() == (("z",), (30,))

    # String representation
    readable = (
        "ExtraDimensions(extra_dim={'z': {'name': 'z', 'values': [5, 10, 15, "
        "20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, "
        "105, 110, 115, 120, 125, 130, 135, 140, 145, 150], 'dtype': "
        "'float64'}}, dim_slice={'z': (0, 30)} coords={'z': <xarray.DataArray "
        "'z' (z: 30)>\narray([  5.,  10.,  15.,  20.,  25.,  30.,  35.,  "
        "40.,  45.,  50.,  55.,\n        60.,  65.,  70.,  75.,  80.,  85.,  "
        "90.,  95., 100., 105., 110.,\n       115., 120., 125., 130., 135., "
        "140., 145., 150.])\nCoordinates:\n  * z        (z) int64 5 10 15 20 "
        "25 30 35 40 ... 120 125 130 135 140 145 150} )"
    )
    assert str(dt.extra_dimensions) == readable
    assert f"{dt.extra_dimensions!r}" == readable


def test_extra_dimensions_exceptions(eo3_metadata, cover_z_dataset_type):
    """Test exceptions on invalid input."""
    dt = DatasetType(eo3_metadata, cover_z_dataset_type)

    # Unknown keys
    with pytest.raises(KeyError) as exc_info:
        dt.extra_dimensions[{"x": (5, 12)}]
    # For some reason the exception message has double quotes around it
    assert str(exc_info.value).strip('"') == "Found unknown keys {'x'} in dim_slices"

    # Bogus measurements_value
    with pytest.raises(ValueError) as exc_info:
        dt.extra_dimensions.measurements_values("x")
    assert str(exc_info.value) == "Dimension x not found."

    # Bogus measurements_slice
    with pytest.raises(ValueError) as exc_info:
        dt.extra_dimensions.measurements_slice("x")
    assert str(exc_info.value) == "Dimension x not found."

    # Bogus measurements_index
    with pytest.raises(ValueError) as exc_info:
        dt.extra_dimensions.measurements_index("x")
    assert str(exc_info.value) == "Dimension x not found."

    # Bogus measurements_index
    with pytest.raises(ValueError) as exc_info:
        dt.extra_dimensions.index_of("x", 50)
    assert str(exc_info.value) == "Dimension x not found."



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
""" reader
"""
from typing import (
    List, Optional, Union, Any, Iterable,
    Tuple, NamedTuple, TypeVar
)
import numpy as np
from affine import Affine
from concurrent.futures import ThreadPoolExecutor
import rasterio                         # type: ignore[import]
from rasterio.io import DatasetReader   # type: ignore[import]
import rasterio.crs                     # type: ignore[import]

from datacube.storage import BandInfo
from datacube.utils.geometry import CRS
from datacube.utils import (
    uri_to_local_path,
    get_part_from_uri,
)
from datacube.drivers._types import (
    ReaderDriverEntry,
    ReaderDriver,
    GeoRasterReader,
    FutureGeoRasterReader,
    FutureNdarray,
    RasterShape,
    RasterWindow,
)

Overrides = NamedTuple('Overrides', [('crs', Optional[CRS]),
                                     ('transform', Optional[Affine]),
                                     ('nodata', Optional[Union[float, int]])])

RioWindow = Tuple[Tuple[int, int], Tuple[int, int]]  # pylint: disable=invalid-name
T = TypeVar('T')


def pick(a: Optional[T], b: Optional[T]) -> Optional[T]:
    """ Return first non-None value or None if all are None
    """
    return b if a is None else a


def _is_netcdf(fmt: str) -> bool:
    return fmt == 'NetCDF'


def _roi_to_window(roi: Optional[RasterWindow], shape: RasterShape) -> Optional[RioWindow]:
    if roi is None:
        return None

    def s2t(s: slice, n: int) -> Tuple[int, int]:
        _in = 0 if s.start is None else s.start
        _out = n if s.stop is None else s.stop

        if _in < 0:
            _in += n
        if _out < 0:
            _out += n

        return (_in, _out)

    s1, s2 = (s2t(s, n)
              for s, n in zip(roi, shape))
    return (s1, s2)


def _dc_crs(crs: Optional[rasterio.crs.CRS]) -> Optional[CRS]:
    """ Convert RIO version of CRS to datacube
    """
    if crs is None:
        return None

    if not crs.is_valid:
        return None

    if crs.is_epsg_code:
        return CRS('epsg:{}'.format(crs.to_epsg()))
    return CRS(crs.wkt)


def _read(src: DatasetReader,
          bidx: int,
          window: Optional[RasterWindow],
          out_shape: Optional[RasterShape]) -> np.ndarray:
    return src.read(bidx,
                    window=_roi_to_window(window, src.shape),
                    out_shape=out_shape)


def _rio_uri(band: BandInfo) -> str:
    """
    - file uris are converted to file names
       - if also netcdf wrap in NETCDF:"${filename}":${layer}
    - All other protocols go through unmodified
    """
    if band.uri_scheme == 'file':
        fname = str(uri_to_local_path(band.uri))

        if _is_netcdf(band.format):
            fname = 'NETCDF:"{}":{}'.format(fname, band.layer)

        return fname

    return band.uri


def _rio_band_idx(band: BandInfo, src: DatasetReader) -> int:
    if band.band is not None:
        return band.band

    if not _is_netcdf(band.format):
        return 1

    bidx = get_part_from_uri(band.uri)
    if bidx is not None:
        return bidx

    if src.count == 1:  # Single-slice netcdf file
        return 1

    raise DeprecationWarning("Stacked netcdf without explicit time index is not supported anymore")


class RIOReader(GeoRasterReader):
    def __init__(self,
                 src: DatasetReader,
                 band_idx: int,
                 pool: ThreadPoolExecutor,
                 overrides: Overrides = Overrides(None, None, None)):

        transform = pick(overrides.transform, src.transform)
        if transform is not None and transform.is_identity:
            transform = None

        self._src = src
        self._crs = overrides.crs or _dc_crs(src.crs)
        self._transform = transform
        self._nodata = pick(overrides.nodata, src.nodatavals[band_idx-1])
        self._band_idx = band_idx
        self._dtype = src.dtypes[band_idx-1]
        self._pool = pool

    @property
    def crs(self) -> Optional[CRS]:
        return self._crs

    @property
    def transform(self) -> Optional[Affine]:
        return self._transform

    @property
    def dtype(self) -> np.dtype:
        return np.dtype(self._dtype)

    @property
    def shape(self) -> RasterShape:
        return self._src.shape

    @property
    def nodata(self) -> Optional[Union[int, float]]:
        return self._nodata

    def read(self,
             window: Optional[RasterWindow] = None,
             out_shape: Optional[RasterShape] = None) -> FutureNdarray:
        return self._pool.submit(_read, self._src, self._band_idx, window, out_shape)


def _compute_overrides(src: DatasetReader, bi: BandInfo) -> Overrides:
    """ If dataset is missing nodata, crs or transform.
    """
    crs, transform, nodata = None, None, None

    if src.crs is None or not src.crs.is_valid:
        crs = bi.crs

    if src.transform.is_identity:
        transform = bi.transform

    if src.nodata is None:
        nodata = bi.nodata

    return Overrides(crs=crs, transform=transform, nodata=nodata)


def _rdr_open(band: BandInfo, ctx: Any, pool: ThreadPoolExecutor) -> RIOReader:
    """ Open file pointed by BandInfo and return RIOReader instance.

        raises Exception on failure

        TODO: start using ctx for handle cache
    """
    normalised_uri = _rio_uri(band)
    src = rasterio.open(normalised_uri, 'r')
    bidx = _rio_band_idx(band, src)

    return RIOReader(src, bidx, pool, _compute_overrides(src, band))


class RIORdrDriver(ReaderDriver):
    def __init__(self, pool: ThreadPoolExecutor, cfg: dict):
        self._pool = pool
        self._cfg = cfg

    def new_load_context(self,
                         bands: Iterable[BandInfo],
                         old_ctx: Optional[Any]) -> Any:
        return None  # TODO: implement file handle cache with this

    def open(self, band: BandInfo, ctx: Any) -> FutureGeoRasterReader:
        return self._pool.submit(_rdr_open, band, ctx, self._pool)


class RDEntry(ReaderDriverEntry):
    PROTOCOLS = ['file', 'http', 'https', 's3', 'ftp', 'zip']
    FORMATS = ['GeoTIFF', 'NetCDF', 'JPEG2000']

    @property
    def protocols(self) -> List[str]:
        return RDEntry.PROTOCOLS

    @property
    def formats(self) -> List[str]:
        return RDEntry.FORMATS

    def supports(self, protocol: str, fmt: str) -> bool:
        # TODO: might need better support matrix structures

        if fmt == 'NetCDF':
            return protocol == 'file'

        return True

    def new_instance(self, cfg: dict) -> ReaderDriver:
        cfg = cfg.copy()
        pool = cfg.pop('pool', None)
        if pool is None:
            max_workers = cfg.pop('max_workers', 1)
            pool = ThreadPoolExecutor(max_workers=max_workers)
        elif not isinstance(pool, ThreadPoolExecutor):
            if not cfg.pop('allow_custom_pool', False):
                raise ValueError("External `pool` should be a `ThreadPoolExecutor`")

        return RIORdrDriver(pool, cfg)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest

from integration_tests.utils import prepare_test_ingestion_configuration
from integration_tests.test_end_to_end import PROJECT_ROOT

COMPLIANCE_CHECKER_NORMAL_LIMIT = 2


@pytest.mark.timeout(20)
@pytest.mark.usefixtures('default_metadata_type',
                         'indexed_ls5_scene_products')
def test_invalid_ingestor_config(clirunner, index, tmpdir):
    """
    Test that the ingestor correctly identifies an invalid ingestor config file.

    Note: We do not need to test valid config files as that is covered by the existing
          ingestor tests.
    """
    base = PROJECT_ROOT / 'integration_tests/data/ingester/'

    for cfg, err in (('invalid_config.yaml', "'src_varname' is a required property"),
                     ('invalid_src_name.yaml', 'No such variable in the source product:')):
        config = base / cfg
        config_path, config = prepare_test_ingestion_configuration(tmpdir, None, config)

        result = clirunner(['ingest', '--config-file', str(config_path)],
                           expect_success=False)

        assert result.exit_code != 0
        assert err in result.output



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Tests for MultiprocessingExecutor
"""

from datacube.executor import get_executor
from time import sleep
import pytest

DATA = [1, 2, 3, 4]
RETRIES = 5


def _echo(x, please_fail=False):
    if please_fail or x == 'please fail':
        raise IOError('Fake I/O error, because you asked')
    return x


def run_executor_tests(executor, sleep_time=1):
    # get_ready: mostly pending
    futures = executor.map(_echo, DATA)
    assert len(futures) == len(DATA)
    completed, failed, pending = executor.get_ready(futures)
    assert len(failed) == 0
    assert len(completed) + len(pending) == len(DATA)

    # get_ready: processed + failure
    data = ['please fail'] + DATA
    futures = executor.map(_echo, data)
    assert len(futures) == len(data)

    for _ in range(RETRIES):
        sleep(sleep_time)  # give it "enough" time to finish
        completed, failed, pending = executor.get_ready(futures)
        if len(pending) == 0:
            break

    if sleep_time:
        assert len(failed) == 1
    else:
        assert len(failed) in [0, 1]

    assert len(completed) + len(pending) + len(failed) == len(data)

    # test results
    futures = executor.map(_echo, DATA)
    results = executor.results(futures)

    assert len(results) == len(DATA)
    assert set(results) == set(DATA)

    # Test failure pass-through
    future = executor.submit(_echo, "", please_fail=True)

    for ff in executor.as_completed([future]):
        with pytest.raises(IOError):
            executor.result(ff)

    # Next completed with data
    future = executor.submit(_echo, 'tt')
    futures = [future]
    result, futures = executor.next_completed(futures, 'default')
    assert len(futures) == 0
    print(type(result), result)

    # Next completed with empty list
    result, futures = executor.next_completed([], 'default')
    assert result == 'default'
    assert len(futures) == 0

    executor.release(future)


def test_concurrent_executor():
    executor = get_executor(None, 2)
    assert 'Multiproc' in str(executor)
    run_executor_tests(executor)

    executor = get_executor(None, 2, use_cloud_pickle=False)
    assert 'Multiproc' in str(executor)
    run_executor_tests(executor)


def test_fallback_executor():
    executor = get_executor(None, None)
    assert 'Serial' in str(executor)

    run_executor_tests(executor, sleep_time=0)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import Mapping, Any, cast
import copy

from .impl import VirtualProduct, Transformation, VirtualProductException
from .impl import from_validated_recipe, virtual_product_kind
from .transformations import MakeMask, ApplyMask, ToFloat, Rename, Select, Expressions
from .transformations import XarrayReduction, year, month, week, day, earliest_time, fiscal_year
from .catalog import Catalog
from .utils import reject_keys

from datacube.model import Measurement
from datacube.utils import import_function
from datacube.utils.documents import parse_yaml

__all__ = ['construct', 'Transformation', 'Measurement']


class NameResolver:
    """ Apply a mapping from name to callable objects in a recipe. """

    def __init__(self, lookup_table):
        self.lookup_table = lookup_table

    def clone(self):
        """ Safely copy the resolver in order to possibly extend it. """
        return NameResolver(copy.deepcopy(self.lookup_table))

    @staticmethod
    def _assert(cond, msg):
        if not cond:
            raise VirtualProductException(msg)

    def construct(self, **recipe) -> VirtualProduct:
        """ Validate recipe and construct virtual product. """

        get = recipe.get

        def lookup(name, namespace=None, kind='transformation'):
            if callable(name):
                return name

            if namespace is not None and namespace in self.lookup_table and name in self.lookup_table[namespace]:
                result = self.lookup_table[namespace][name]
            else:
                try:
                    result = import_function(name)
                except (ImportError, AttributeError):
                    msg = "could not resolve {} {} in {}".format(kind, name, recipe)
                    raise VirtualProductException(msg)

            self._assert(callable(result), "{} not callable in {}".format(kind, recipe))

            return result

        kind = virtual_product_kind(recipe)

        if kind == 'product':
            func_keys = ['fuse_func', 'dataset_predicate']
            return from_validated_recipe({key: value if key not in func_keys else lookup(value, kind='function')
                                          for key, value in recipe.items()})

        if kind == 'transform':
            cls_name = recipe['transform']
            input_product = cast(Mapping, get('input'))

            self._assert(input_product is not None, "no input for transformation in {}".format(recipe))

            return from_validated_recipe(dict(transform=lookup(cls_name, 'transform'),
                                              input=self.construct(**input_product),
                                              **reject_keys(recipe, ['transform', 'input'])))

        if kind == 'collate':
            self._assert(len(recipe['collate']) > 0, "no children for collate in {}".format(recipe))

            return from_validated_recipe(dict(collate=[self.construct(**child) for child in recipe['collate']],
                                              **reject_keys(recipe, ['collate'])))

        if kind == 'juxtapose':
            self._assert(len(recipe['juxtapose']) > 0, "no children for juxtapose in {}".format(recipe))

            return from_validated_recipe(dict(juxtapose=[self.construct(**child) for child in recipe['juxtapose']],
                                              **reject_keys(recipe, ['juxtapose'])))

        if kind == 'aggregate':
            cls_name = recipe['aggregate']
            input_product = cast(Mapping, get('input'))
            group_by = get('group_by')

            self._assert(input_product is not None, "no input for aggregate in {}".format(recipe))
            self._assert(group_by is not None, "no group_by for aggregate in {}".format(recipe))

            return from_validated_recipe(dict(aggregate=lookup(cls_name, 'aggregate'),
                                              group_by=lookup(group_by, 'aggregate/group_by', kind='group_by'),
                                              input=self.construct(**input_product),
                                              **reject_keys(recipe, ['aggregate', 'input', 'group_by'])))

        if kind == 'reproject':
            input_product = cast(Mapping, get('input'))
            output_crs = recipe['reproject'].get('output_crs')
            resolution = recipe['reproject'].get('resolution')
            align = recipe['reproject'].get('align')

            self._assert(input_product is not None, "no input for reproject in {}".format(recipe))
            self._assert(output_crs is not None, "no output_crs for reproject in {}".format(recipe))
            self._assert(resolution is not None, "no resolution for reproject in {}".format(recipe))

            return from_validated_recipe(dict(reproject=recipe['reproject'],
                                              input=self.construct(**input_product),
                                              **reject_keys(recipe, ['reproject', 'input'])))

        raise VirtualProductException("could not understand virtual product recipe: {}".format(recipe))

    def register(self, namespace: str, name: str, callable_obj):
        """ Register a callable to the name resolver. """
        if namespace not in self.lookup_table:
            self.lookup_table[namespace] = {}

        if name in self.lookup_table[namespace]:
            raise VirtualProductException("name {} under {} is already registered".format(name, namespace))

        self.lookup_table[namespace][name] = callable_obj


DEFAULT_RESOLVER = NameResolver({'transform': dict(make_mask=MakeMask,
                                                   apply_mask=ApplyMask,
                                                   to_float=ToFloat,
                                                   rename=Rename,
                                                   select=Select,
                                                   expressions=Expressions),
                                 'aggregate': dict(xarray_reduction=XarrayReduction),
                                 'aggregate/group_by': dict(year=year,
                                                            month=month,
                                                            week=week,
                                                            day=day,
                                                            earliest_time=earliest_time,
                                                            fiscal_year=fiscal_year)})


def construct(name_resolver=None, **recipe: Mapping[str, Any]) -> VirtualProduct:
    """
    Create a virtual product from a specification dictionary.
    """
    if name_resolver is None:
        name_resolver = DEFAULT_RESOLVER

    return name_resolver.construct(**recipe)


def construct_from_yaml(recipe: str, name_resolver=None) -> VirtualProduct:
    """
    Create a virtual product from a yaml recipe.
    """
    if name_resolver is None:
        name_resolver = DEFAULT_RESOLVER

    return construct(name_resolver=name_resolver, **parse_yaml(recipe))


def catalog_from_yaml(catalog_body: str, name_resolver=None) -> Catalog:
    """
    Load a catalog of virtual products from a yaml document.
    """
    if name_resolver is None:
        name_resolver = DEFAULT_RESOLVER

    return Catalog(name_resolver, parse_yaml(catalog_body))


def catalog_from_file(filename: str, name_resolver=None) -> Catalog:
    """
    Load a catalog of virtual products from a yaml file.
    """
    with open(filename) as fl:
        return catalog_from_yaml(fl.read(), name_resolver=name_resolver)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Tables for indexing the datasets which were ingested into the AGDC.
"""

import logging
from typing import Type

from sqlalchemy.orm import aliased, registry, relationship, column_property
from sqlalchemy import ForeignKey, UniqueConstraint, PrimaryKeyConstraint, CheckConstraint, SmallInteger, Text
from sqlalchemy import Column, Integer, String, DateTime
from sqlalchemy.dialects import postgresql as postgres
from sqlalchemy.sql import func

from . import sql
from . import _core

_LOG = logging.getLogger(__name__)

orm_registry = registry()


@orm_registry.mapped
class MetadataType:
    __tablename__ = "metadata_type"
    __table_args__ = (
        _core.METADATA,
        CheckConstraint(r"name ~* '^\w+$'", name='alphanumeric_name'),
        {
            "schema": sql.SCHEMA_NAME,
            "comment": "Metadata type, defining search fields requiring dynamic indexes",
        }
    )
    id = Column(SmallInteger, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=True, comment="A human-friendly name/label for this metadata type")
    definition = Column(postgres.JSONB, nullable=False, comment="metadata schema with search fields")
    # When it was added and by whom.
    added = Column(DateTime(timezone=True), server_default=func.now(), nullable=False, comment="when added")
    added_by = Column(Text, server_default=func.current_user(), nullable=False, comment="added by whom")

    products = relationship("Product")
    datasets = relationship("Dataset")


@orm_registry.mapped
class Product:
    __tablename__ = "product"
    __table_args__ = (
        _core.METADATA,
        CheckConstraint(r"name ~* '^\w+$'", name='alphanumeric_name'),
        {
            "schema": sql.SCHEMA_NAME,
            "comment": "A product or dataset type, family of related datasets."
        }
    )
    id = Column(SmallInteger, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=False, comment="A human-friendly name/label for this product")
    # DB column named metadata for (temporary) backwards compatibility,
    # but is forbidden by SQLAlchemy declarative style
    metadata_doc = Column(name="metadata",
                          type_=postgres.JSONB, nullable=False,
                          comment="""The product metadata document (subset of the full definition)
All datasets of this type should contain these fields.
(newly-ingested datasets may be matched against these fields to determine the dataset type)""")
    metadata_type_ref = Column(SmallInteger, ForeignKey(MetadataType.id), nullable=False,
                               comment="The metadata type - how to interpret the metadata")
    definition = Column('definition', postgres.JSONB, nullable=False, comment="Full product definition document")
    added = Column(DateTime(timezone=True), server_default=func.now(), nullable=False, comment="when added")
    added_by = Column(Text, server_default=func.current_user(), nullable=False, comment="added by whom")

    datasets = relationship("Dataset")


@orm_registry.mapped
class Dataset:
    __tablename__ = "dataset"
    __table_args__ = (
        _core.METADATA,
        {
            "schema": sql.SCHEMA_NAME,
            "comment": "A dataset."
        }
    )
    id = Column(postgres.UUID(as_uuid=True), primary_key=True)
    metadata_type_ref = Column(SmallInteger, ForeignKey(MetadataType.id), nullable=False,
                               comment="The metadata type - how to interpret the metadata")
    product_ref = Column(SmallInteger, ForeignKey(Product.id), nullable=False,
                         comment="The product this dataset belongs to")
    # DB column named metadata for (temporary) backwards compatibility,
    # but is forbidden by SQLAlchemy declarative style
    metadata_doc = Column(name="metadata", type_=postgres.JSONB, index=False, nullable=False,
                          comment="The dataset metadata document")
    archived = Column(DateTime(timezone=True), default=None, nullable=True,
                      comment="when archived, null if active")
    added = Column(DateTime(timezone=True), server_default=func.now(), nullable=False, comment="when added")
    added_by = Column(Text, server_default=func.current_user(), nullable=False, comment="added by whom")

    locations = relationship("DatasetLocation", viewonly=True)
    active_locations = relationship("DatasetLocation",
                                    primaryjoin="and_(Dataset.id==DatasetLocation.dataset_ref, "
                                                "DatasetLocation.archived==None)",
                                    viewonly=True,
                                    order_by="desc(DatasetLocation.added)")
    archived_locations = relationship("DatasetLocation",
                                      viewonly=True,
                                      primaryjoin="and_(Dataset.id==DatasetLocation.dataset_ref, "
                                                  "DatasetLocation.archived!=None)"
                                     )


@orm_registry.mapped
class DatasetLocation:
    __tablename__ = "location"
    __table_args__ = (
        _core.METADATA,
        UniqueConstraint('uri_scheme', 'uri_body', 'dataset_ref'),
        {
            "schema": sql.SCHEMA_NAME,
            "comment": "Where data for the dataset can be found (uri)."
        }
    )
    id = Column(Integer, primary_key=True, autoincrement=True)
    dataset_ref = Column(postgres.UUID(as_uuid=True), ForeignKey(Dataset.id), nullable=False,
                         comment="The product this dataset belongs to")
    uri_scheme = Column(String, nullable=False, comment="The scheme of the uri.")
    uri_body = Column(String, nullable=False, comment="""The body of the uri.

The uri scheme and body make up the base URI to find the dataset.

All paths in the dataset metadata can be computed relative to this.
(it is often the path of the source metadata file)

eg 'file:///g/data/datasets/LS8_NBAR/odc-metadata.yaml' or 'ftp://eo.something.com/dataset'
'file' is a scheme, '///g/data/datasets/LS8_NBAR/odc-metadata.yaml' is a body.""")
    added = Column(DateTime(timezone=True), server_default=func.now(), nullable=False, comment="when added")
    added_by = Column(Text, server_default=func.current_user(), nullable=False, comment="added by whom")
    archived = Column(DateTime(timezone=True), default=None, nullable=True,
                      comment="when archived, null for the active location")
    uri = column_property(uri_scheme + ':' + uri_body)
    dataset = relationship("Dataset")


SelectedDatasetLocation = aliased(DatasetLocation, name="sel_loc")


@orm_registry.mapped
class DatasetSource:
    __tablename__ = "dataset_lineage"
    __table_args__ = (
        _core.METADATA,
        PrimaryKeyConstraint('dataset_ref', 'classifier'),
        UniqueConstraint('source_dataset_ref', 'dataset_ref'),
        {
            "schema": sql.SCHEMA_NAME,
            "comment": "Represents a source-lineage relationship between two datasets"
        }
    )
    dataset_ref = Column(postgres.UUID(as_uuid=True), nullable=False, index=True,
                         comment="The downstream derived dataset produced from the upstream source dataset.")
    source_dataset_ref = Column(
        postgres.UUID(as_uuid=True), nullable=False, index=True,
        comment="An upstream source dataset that the downstream derived dataset was produced from."
    )
    classifier = Column(String, nullable=False, comment="""An identifier for this source dataset.
E.g. the dataset type ('ortho', 'nbar'...) if there's only one source of each type, or a datestamp
for a time-range summary.""")


class SpatialIndex:
    """
    Base class for dynamically SpatialIndex ORM models (See _spatial.py)
    """


@orm_registry.mapped
class SpatialIndexRecord:
    __tablename__ = "spatial_indicies"
    __table_args__ = (
        _core.METADATA,
        {
            "schema": sql.SCHEMA_NAME,
            "comment": "Record of the existence of a Spatial Index Table for an SRID/CRS",
        }
    )
    srid = Column(SmallInteger, primary_key=True, autoincrement=False)
    table_name = Column(String,
                        unique=True, nullable=True,
                        comment="The name of the table implementing the index - DO NOT CHANGE")
    added = Column(DateTime(timezone=True), server_default=func.now(), nullable=False, comment="when added")
    added_by = Column(Text, server_default=func.current_user(), nullable=False, comment="added by whom")

    @classmethod
    def from_spindex(cls, spindex: Type[SpatialIndex]) -> "SpatialIndexRecord":
        return cls(srid=spindex.__tablename__[8:],
                   table_name=spindex.__tablename__)


ALL_STATIC_TABLES = [
    MetadataType.__table__, Product.__table__, Dataset.__table__,
    DatasetLocation.__table__, DatasetSource.__table__, SpatialIndexRecord.__table__
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Access methods for indexing datasets & products.
"""

import logging

from datacube.config import LocalConfig
from datacube.index.abstract import AbstractIndex

_LOG = logging.getLogger(__name__)


def index_connect(local_config: LocalConfig = None,
                  application_name: str = None,
                  validate_connection: bool = True) -> AbstractIndex:
    """
    Create a Data Cube Index (as per config)

    It contains all the required connection parameters, but doesn't actually
    check that the server is available.

    :param application_name: A short, alphanumeric name to identify this application.
    :param local_config: Config object to use. (optional)
    :param validate_connection: Validate database connection and schema immediately
    :raises datacube.index.Exceptions.IndexSetupError:
    """
    from datacube.drivers import index_driver_by_name, index_drivers

    if local_config is None:
        local_config = LocalConfig.find()

    driver_name = local_config.get('index_driver', 'default')
    index_driver = index_driver_by_name(driver_name)
    if not index_driver:
        raise RuntimeError(
            "No index driver found for %r. %s available: %s" % (
                driver_name, len(index_drivers()), ', '.join(index_drivers())
            )
        )

    return index_driver.connect_to_index(local_config,
                                         application_name=application_name,
                                         validate_connection=validate_connection)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0


def fail_to_init():
    raise ValueError('This will always fail')


def init_to_none():
    return None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from datacube.utils.masking import *  # noqa: F401, F403

import warnings

warnings.warn("datacube.storage.masking has moved to datacube.utils.masking",
              category=DeprecationWarning)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from dateutil.tz import tzutc

from datacube.model import Range
from datacube.ui import parse_expressions


def test_parse_empty_str():
    q = parse_expressions('')
    assert q == {}


def test_id_search():
    q = parse_expressions("id = 26931d17-7a4e-4b55-98e7-d6777fb61df6")
    assert q['id'] == "26931d17-7a4e-4b55-98e7-d6777fb61df6"

    q = parse_expressions('id = "26931d17-7a4e-4b55-98e7-d6777fb61df6"')
    assert q['id'] == "26931d17-7a4e-4b55-98e7-d6777fb61df6"


def test_simple_string():
    q = parse_expressions("region_code = 56KKD", "i=10", "f=10.3")
    assert q['region_code'] == '56KKD'
    assert q['i'] == 10
    assert q['f'] == 10.3


def test_between_expression():
    q = parse_expressions('time in [2014, 2015]')
    assert 'time' in q
    r = q['time']
    assert isinstance(r, Range)
    assert isinstance(r.begin, datetime)
    assert isinstance(r.end, datetime)

    for k in ('lon', 'lat', 'x', 'y'):
        q = parse_expressions('{} in [10, 11.3]'.format(k))
        assert k in q
        r = q[k]
        assert isinstance(r, Range)
        assert isinstance(r.begin, (int, float))
        assert isinstance(r.end, (int, float))
        assert r == Range(10, 11.3)


def test_parse_simple_expression():
    assert parse_expressions('platform = 4') == {'platform': 4}
    assert parse_expressions('platform = "LANDSAT_8"') == {'platform': 'LANDSAT_8'}
    assert parse_expressions('platform = LANDSAT_8') == {'platform': 'LANDSAT_8'}
    assert parse_expressions('platform = "LAND SAT_8"') == {'platform': 'LAND SAT_8'}

    assert parse_expressions('lat in [4, 6]') == {'lat': Range(4, 6)}


def test_parse_uri_expression():
    assert parse_expressions('uri = file:///f/data/test.nc') == {'uri': 'file:///f/data/test.nc'}
    assert parse_expressions('uri = "file:///f/data/test.nc"') == {'uri': 'file:///f/data/test.nc'}
    assert parse_expressions('uri = "file:///f/data/test me.nc"') == {'uri': 'file:///f/data/test me.nc'}
    assert parse_expressions('uri = file:///C:/f/data/test.nc') == {'uri': 'file:///C:/f/data/test.nc'}
    assert parse_expressions('uri = "file:///C:/f/data/test.nc"') == {'uri': 'file:///C:/f/data/test.nc'}
    assert parse_expressions('uri = "file:///C:/f/data/test me.nc"') == {'uri': 'file:///C:/f/data/test me.nc'}


def test_parse_dates():
    assert parse_expressions('time in 2014-03-02') == {
        'time': Range(begin=datetime(2014, 3, 2, 0, 0, tzinfo=tzutc()),
                      end=datetime(2014, 3, 2, 23, 59, 59, 999999, tzinfo=tzutc()))
    }

    assert parse_expressions('time in 2014-3-2') == {
        'time': Range(begin=datetime(2014, 3, 2, 0, 0, tzinfo=tzutc()),
                      end=datetime(2014, 3, 2, 23, 59, 59, 999999, tzinfo=tzutc()))
    }

    # A missing day defaults to the first of the month.
    # They are probably better off using in-expessions in these cases (eg. "time in 2013-01"), but it's here
    # for backwards compatibility.
    march_2014 = {
        'time': Range(begin=datetime(2014, 3, 1, 0, 0, tzinfo=tzutc()),
                      end=datetime(2014, 3, 31, 23, 59, 59, 999999, tzinfo=tzutc()))
    }
    assert parse_expressions('time in 2014-03') == march_2014
    assert parse_expressions('time in 2014-3') == march_2014

    implied_feb_march_2014 = {
        'time': Range(begin=datetime(2014, 2, 1, 0, 0, tzinfo=tzutc()),
                      end=datetime(2014, 3, 31, 23, 59, 59, 999999, tzinfo=tzutc()))
    }
    assert parse_expressions('time in [2014-02, 2014-03]') == implied_feb_march_2014


def test_parse_date_ranges():
    eighth_march_2014 = {
        'time': Range(datetime(2014, 3, 8, tzinfo=tzutc()), datetime(2014, 3, 8, 23, 59, 59, 999999, tzinfo=tzutc()))
    }
    assert parse_expressions('time in 2014-03-08') == eighth_march_2014
    assert parse_expressions('time in 2014-03-8') == eighth_march_2014

    march_2014 = {
        'time': Range(datetime(2014, 3, 1, tzinfo=tzutc()), datetime(2014, 3, 31, 23, 59, 59, 999999, tzinfo=tzutc()))
    }
    assert parse_expressions('time in 2014-03') == march_2014
    assert parse_expressions('time in 2014-3') == march_2014
    # Leap year, 28 days
    feb_2014 = {
        'time': Range(datetime(2014, 2, 1, tzinfo=tzutc()), datetime(2014, 2, 28, 23, 59, 59, 999999, tzinfo=tzutc()))
    }
    assert parse_expressions('time in 2014-02') == feb_2014
    assert parse_expressions('time in 2014-2') == feb_2014

    # Entire year
    year_2014 = {
        'time': Range(datetime(2014, 1, 1, tzinfo=tzutc()), datetime(2014, 12, 31, 23, 59, 59, 999999, tzinfo=tzutc()))
    }
    assert parse_expressions('time in 2014') == year_2014


def test_parse_multiple_simple_expressions():
    # Multiple expressions in one command-line statement.
    # Mixed whitespace:
    between_exp = parse_expressions('platform=LS8 lat in [-4, 23.5] instrument="OTHER"')
    assert between_exp == {'platform': 'LS8', 'lat': Range(-4, 23.5), 'instrument': 'OTHER'}
    # Range(x,y) is "equal" to (x, y). Check explicitly that it's a range:
    assert between_exp['lat'].begin == -4



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Dict, Any, Tuple, Iterable

_LOG = logging.getLogger(__name__)


def load_drivers(group: str) -> Dict[str, Any]:
    """
    Load available drivers for a given group name.

    Gracefully handles:

     - Driver module not able to be imported
     - Driver init function throwing an exception or returning None

     By having driver entry_points pointing to a function, we defer loading the driver
     module or running any code until required.

    :param group: Name of the entry point group e.g. "datacube.plugins.io.read"

    :returns: Dictionary String -> Driver Object
    """

    def safe_load(ep):
        from pkg_resources import DistributionNotFound
        # pylint: disable=broad-except,bare-except
        try:
            driver_init = ep.load()
        except DistributionNotFound:
            # This happens when entry points were marked with extra features,
            # but extra feature were not requested for installation
            return None
        except Exception as e:
            _LOG.warning('Failed to resolve driver %s::%s', group, ep.name)
            _LOG.warning('Error was: %s', repr(e))
            return None

        try:
            driver = driver_init()
        except Exception:
            _LOG.warning('Exception during driver init, driver name: %s::%s', group, ep.name)
            return None

        if driver is None:
            _LOG.warning('Driver init returned None, driver name: %s::%s', group, ep.name)

        return driver

    def resolve_all(group: str) -> Iterable[Tuple[str, Any]]:
        from pkg_resources import iter_entry_points
        for ep in iter_entry_points(group=group, name=None):
            driver = safe_load(ep)
            if driver is not None:
                yield (ep.name, driver)

    return dict((name, driver) for name, driver in resolve_all(group))
